package pltag.german;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;


public class PredictionTree extends ElementaryTree {
	
	private int index;
	
	// in case this is a pre-combined prediction tree, the individual prediction trees
	// are stored here
	private ArrayList<ElementaryTree> individuals = new ArrayList<ElementaryTree>();
	
	public PredictionTree(int i) {
		index = i;
	}

	
	public PredictionTree(int i, ElementaryTree eTree) {
		index = i;
		setType(eTree.getType());
		// assume that the left-most anchor is the point of recognition for
		// collocations/MATs.
		ElementaryTreeNode anch = eTree.getLeftMostAnchor();
		ArrayList<ElementaryTreeNode> trunk = anch.getPathUpwards();
		turnElementaryTreeIntoPredTree(eTree.getRoot(), eTree, trunk);
		// if this is a left adjoining (or wrapping) auxiliary tree,
		// the foot node is right of the spine, so it was not added
		// in the preceding procedure. we do this now:
		if (getType() == NodeStatus.ADJ && getFoot() == null) {
			addFootPath(eTree.getFoot(), eTree);
		}
		
		if (getType() == NodeStatus.MOD) {
			setModParent(eTree.getModParent());
		}
		individuals.add(this);
		
	}



	/*
	 * Precondition: The compTree is a combination of PredictionTrees
	 */
	public PredictionTree(int i, CompositeTree compTree) {
		index = i;
		setType(compTree.getType());
		copy(compTree.getRoot(), compTree);
		if (getType() == NodeStatus.MOD) {
			setModParent(compTree.getModParent());
		}
		individuals.addAll(compTree.getIndividualTrees());
	}


	public int getIndex() {
		return index;
	}
	
	public void setIndex(int i) {
		index = i;
	}
	
	public ArrayList<ElementaryTree> getIndividualPredictionTrees() {
		return individuals;
	}
	

	private ElementaryTreeNode copy(ElementaryTreeNode node, CompositeTree compTree) {
		ElementaryTreeNode nodeCopy = new ElementaryTreeNode(node);
		
		if (node == compTree.getRoot()) {
			setRoot(nodeCopy);
		}
		if (getType() == NodeStatus.ADJ && node == compTree.getFoot()) {
			setFoot(nodeCopy);
		}
		addNode(nodeCopy);
		
		for (ElementaryTreeNode child : node.getChildren()) {
			ElementaryTreeNode childCopy = copy(child, compTree);
			nodeCopy.addChild(childCopy);
			childCopy.setParent(nodeCopy);
		}
		return nodeCopy;
	}

	
	private ElementaryTreeNode turnElementaryTreeIntoPredTree(ElementaryTreeNode node, 
			ElementaryTree eTree, ArrayList<ElementaryTreeNode> trunk) {
		ElementaryTreeNode nodeCopy = new ElementaryTreeNode(node);
		if (node == eTree.getRoot()) {
			setRoot(nodeCopy);
		}
		if (getType() == NodeStatus.ADJ && node == eTree.getFoot()) {
			setFoot(nodeCopy);
		}
		addNode(nodeCopy);
		
		// set origins
		if (node == eTree.getRoot() && (getType() == NodeStatus.ARG || getType() == NodeStatus.ADJ)) {
			nodeCopy.setOriginBottom(this);
		}
		else if (nodeCopy.isSubst() || node == eTree.getFoot()) {
			nodeCopy.setOriginTop(this);
		}
		else {
			// inner nodes, leaf nodes, root nodes of MOD
			nodeCopy.setOriginTop(this);
			nodeCopy.setOriginBottom(this);
		}
		
		
		if (nodeCopy.isLeaf()) {
			if (!nodeCopy.isTrace()) { // predication trees are non-lexicalized
				nodeCopy.setLeaf("");
				nodeCopy.setLemma("");
			}
		}
		else if (!nodeCopy.isSubst()){
//			ElementaryTreeNode headChild = node.getHeadChild();
//			int idOfHeadChild;
//			if (headChild != null) {
//				idOfHeadChild = node.getHeadChild().getID();
//			}
//			else {
//				idOfHeadChild = nodeCopy.getTreeNodeBottom().getHeadChild().getID();
//			}
			// add all siblings left of the head trunk
			for (ElementaryTreeNode child : node.getChildren()) {
				ElementaryTreeNode childCopy = turnElementaryTreeIntoPredTree(child, eTree, trunk);
				nodeCopy.addChild(childCopy);
				childCopy.setParent(nodeCopy);
				if (child.isHead()) {
					nodeCopy.setHeadChild(childCopy);
					childCopy.setHead();
				}
//				if (child.getID() == idOfHeadChild) {
				if (trunk.contains(child)) {
					// hack
					if (!child.isHead()) {
						nodeCopy.setHeadChild(childCopy);
						childCopy.setHead();
					}
					break;
				}
			}
		}
		return nodeCopy;
	}
	



	private ElementaryTreeNode addFootPath(ElementaryTreeNode node, ElementaryTree eTree) {
		ElementaryTreeNode nodeCopy = new ElementaryTreeNode(node);
		if (eTree.getType() == NodeStatus.ADJ && getType() == NodeStatus.ADJ &&
				node == eTree.getFoot()) {
			setFoot(nodeCopy);
			nodeCopy.setOriginTop(this);
		}
		addNode(nodeCopy);
		
		ElementaryTreeNode parent = node.getParent();
		ElementaryTreeNode parentInPredTree = containsID(parent.getID());
		if (parentInPredTree == null) { // TODO: test!!
			ElementaryTreeNode parentCopy = addFootPath(parent, eTree);
			parentCopy.setOriginTop(this);
			parentCopy.setOriginBottom(this);
			parentCopy.addChild(nodeCopy);
			nodeCopy.setParent(parentCopy);
		}
		else {
			parentInPredTree.addChild(nodeCopy);
			nodeCopy.setParent(parentInPredTree);
		}
		
		return nodeCopy;
	}
	
	// We don't have unary productions that stem from auxiliary trees, as 
	// we already factored them out before
	public void cutUnary(ElementaryTreeNode node, HashSet<TreeNode> nodesRequired) {
		
		if (node.getChildren().size() > 0) {
			cutUnary(node.getHeadChild(), nodesRequired);
		}
		
		if (node.getChildren().size() == 1) {
			ElementaryTreeNode child = node.getChildren().get(0);
			if (!nodesRequired.contains(child.getTreeNodeBottom())
					&& !nodesRequired.contains(child.getTreeNodeTop())
					&& node != getRoot()
					&& child.getChildren().size() == 0) {
				node.getChildren().remove(child);
				getNodes().remove(child);
				child.setParent(null);
			}
//			boolean modToTheLeft = false;
//			TreeNode treeNode = node.getTreeNodeBottom(); 
//			for (TreeNode treeNodeChild : treeNode.getChildren()) {
//				if (treeNodeChild.isHead()) {
//					break;
//				}
//				if (treeNodeChild.getNodeStatus() == NodeStatus.MOD) {
//					modToTheLeft = true;
//					break;
//				}
//			}
//			if (!modToTheLeft) {
//				ElementaryTreeNode child = node.getChildren().get(0);
//				if (node != getRoot() && child.getChildren().size() == 0) {
//					node.getChildren().remove(child);
//					getNodes().remove(child);
//					child.setParent(null);
//				}
//			}
		}
		
	}
	
	
	
	public String toString() {
		return toStringLex(new HashMap<String, Trace>());
	}
	
	public String toStringLex(boolean motherLevelForSisterAdjunction,
			HashMap<String, Trace> traceListTigerTree) {
		setMotherLevelForSisterAdj(motherLevelForSisterAdjunction);
		return toStringLex(traceListTigerTree);
	}
	
	
	public String toStringLex(HashMap<String, Trace> traceListTigerTree) {
		StringBuffer sb = new StringBuffer("prediction:\t");
		sb.append(getType());
		sb.append("\t");
		
		traceMappingForOutput.clear();
		
		// add the mother level to modifier trees 
		// if the motherLevelForSisterAdj option is set accordingly
		if (getMotherLevelForSisterAdj() && getType() == NodeStatus.MOD) {
			int top = individuals.indexOf(getRoot().getOriginTop());
			String topOrigin;
			if (top == -1) {
				if (getRoot().getOriginTop() == this) topOrigin = "x";
				else topOrigin = "null";
			}
			else {
				topOrigin = String.valueOf(top+1);
			}
			
			sb.append("(");
			sb.append(getModParent().getPhrasalCategory());
			if (!getModParent().getFunctionalCategory().equals("--")) {
				sb.append("-");
				sb.append(getModParent().getFunctionalCategory());
			}
			sb.append("^null_").append(topOrigin);
			sb.append("* ");
			
			toStringStruct(getRoot(), sb, traceListTigerTree);
			
			sb.append(")");
		}
		else {
			toStringStruct(getRoot(), sb, traceListTigerTree);
		}
		
		//toStringStruct(getRoot(), sb);
		return sb.toString();
	}

	private void toStringStruct(ElementaryTreeNode node, StringBuffer sb,
			HashMap<String, Trace> traceListTigerTree) {
		int top = individuals.indexOf(node.getOriginTop());
		int bottom = individuals.indexOf(node.getOriginBottom());
		String topOrigin;
		String bottomOrigin;
		if (top == -1) {
			topOrigin = "null";
		}
		else {
			topOrigin = String.valueOf(top+1);
		}
		if (bottom == -1) {
			bottomOrigin = "null";
		}
		else {
			bottomOrigin = String.valueOf(bottom+1);
		}
		
		sb.append("(");
		
		// phrasal category
		if (node.isTrace()) {
			if (!traceMappingForOutput.containsKey(node.getTraceID())) {
				traceMappingForOutput.put(node.getTraceID(), traceMappingForOutput.size()+1);
			}
			sb.append("*T");
			sb.append(traceMappingForOutput.get(node.getTraceID()));
			sb.append("*");
		}
		else {
			sb.append(node.getPhrasalCategory());
		}
		
		// moved further down
//		// filler information
//		if (node.isFiller() && traceListTigerTree.containsKey(node.getTraceID())) {
//			// CHANGE
//			//sb.append("-");
//			//sb.append(node.getTraceID());
//			
//			if (!traceMappingForOutput.containsKey(node.getTraceID())) {
//				traceMappingForOutput.put(node.getTraceID(), traceMappingForOutput.size()+1);
//			}
//			sb.append("{T");
//			sb.append(traceMappingForOutput.get(node.getTraceID()));
//			sb.append("}");
//		}
		
		// functional category
		if (!node.getFunctionalCategory().equals("--")) {
			sb.append("-");
			sb.append(node.getFunctionalCategory());
		}
		
		// morphology
		sb.append(node.getMorphology());
		
		// filler information
		if (node.isFiller() && traceListTigerTree.containsKey(node.getTraceID())) {
			// CHANGE
			//sb.append("-");
			//sb.append(node.getTraceID());
			
			if (!traceMappingForOutput.containsKey(node.getTraceID())) {
				traceMappingForOutput.put(node.getTraceID(), traceMappingForOutput.size()+1);
			}
			sb.append("{T");
			sb.append(traceMappingForOutput.get(node.getTraceID()));
			sb.append("}");
		}
		
		// incoming secondary edges
		if (!bottomOrigin.equals("null")) {
			for (SecondaryEdge sec : node.getTreeNodeBottom().getIncomingSecondaryEdges()) {
				TreeNode dependent = sec.getStartNode();
				NodeStatus ns = dependent.getNodeStatus();
				if (ns == NodeStatus.ARG || ns == NodeStatus.MAARG || dependent.isHead() 
						|| (ns == NodeStatus.MOD && Utils.isAuxiliaryOrModal(dependent))) {
					sb.append("$$").append(dependent.getPhrasalCategory());
					if (!dependent.getFunctionalCategory().equals("--")) {
						sb.append("-").append(dependent.getFunctionalCategory());
					}
					sb.append(dependent.getTop().getElementaryTree().getNodeWithTop(dependent).getMorphology());
					sb.append("$$");
				}
			}
		}
		
		if (node.isSubst()) {
			sb.append("^").append(topOrigin).append("_").append("null");
			sb.append("!");
		}
		else if (getType() == NodeStatus.MOD && node == getRoot()) {
			sb.append("^").append(topOrigin).append("_").append(bottomOrigin);
			if(!getMotherLevelForSisterAdj()) {
				sb.append("*");
			}
		}
		else if (getType() == NodeStatus.ADJ && node == getFoot()) {
			sb.append("^").append(topOrigin).append("_").append("null");
			sb.append("*");
		}
		else if (node == getRoot()) {
			sb.append("^").append("null").append("_").append(bottomOrigin);
		}
		else {
			sb.append("^").append(topOrigin).append("_").append(bottomOrigin);
		}
		
		sb.append(" ");
		
		if (node.isLeaf() && node.isTrace()) {
			sb.append("-");
		}

		for (ElementaryTreeNode child : node.getChildren()) {
			toStringStruct(child, sb, traceListTigerTree);
		}

		
		sb.append(")");
	}


//	private void toString(ElementaryTreeNode node, StringBuffer sb) {
//		sb.append("(");
//		sb.append(node.getFullCategory());
//		
//		if (node.getTop() != null && node.getTop().getLeafLink().getID() > index) {
//			sb.append("^");
//			sb.append(node.getTop().getLeafLink().getID());
//		}
//		else {
//			sb.append("^null");
//		}
//		
//		if (node.getBottom() != null && node.getBottom().getLeafLink().getID() > index) {
//			sb.append("_");
//			sb.append(node.getBottom().getLeafLink().getID());
//		}
//		else {
//			sb.append("_null");
//			
//			if (node.getNodeStatus() == NodeStatus.ARG || 
//					node.getNodeStatus() == NodeStatus.MAARG) {
//				sb.append("!");
//			}
//		}
//		
//		sb.append(" ");
//		for (TreeNode child : node.getChildren()) {
//			if (nodes.contains(child)) {
//				toString(child, sb);
//			}
//		}
//		sb.append(")");
//	}

}
