package pltag.german;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;




public class TigerCorpus {
	
	private HashMap<Integer, String> sentences = new HashMap<Integer, String>();
	private HashMap<Integer, TigerTree> trees = new HashMap<Integer, TigerTree>();
	
	public boolean debug = false;
	
	/*
	 * Requires the input file fileName to 
	 *  * to in Penn format, i.e. with resolved crossing branches, 
	 *  * to in utf-8 encoding
	 *  * have $( and $) replaced with $*LRB* and $*RRB* respectively
	 *  * have ( and ) replace with *LRB* and *RRB* respectively
	 *  * give the node id of the negra format of each node after |||
	 */
	public TigerCorpus(File filePenn, File fileNegra) {
		// Read the bracketed tree structures
		System.out.println("Reading the trees ...");
		read(filePenn);
		//Utils.printMemory("TigerCorpus - After reading strings");
		// Build trees from the bracketed structures
		buildTrees(sentences);
		//Utils.printMemory("TigerCorpus - After building trees");
		// Enrich the trees with lemmata, morphology and secondary edges
		System.out.println("Enriching the trees ...");
		enrichTrees(fileNegra);
		//Utils.printMemory("TigerCorpus - After enriching trees");
	}
	

	public HashMap<Integer, TigerTree> getTrees() {
		return trees;
	}
	
	private void buildTrees(HashMap<Integer, String> sents) {
		for (Map.Entry<Integer, String> sent: sents.entrySet()) {
			String s = sent.getValue();
	//		if (sent.getKey() == 47746) {
	//			System.out.println("test");
	//			}

			//System.out.println(trees.size()+"\t"+s);
			if (! s.equals("")){
				TigerTree tree = new TigerTree(s);
				//System.out.println(tree);
				trees.put(sent.getKey(), tree);
				
			}			
		}		
	}

	private void read(File fileName) {
		StringBuffer sentence = new StringBuffer();
		
		FileReader fr ;
		BufferedReader input = null;
		try {			
			fr = new FileReader(fileName);
			input = new BufferedReader(fr);
			
			String line = null;
			Pattern p = Pattern.compile("%% Sent ([0-9]+)");
			String sentID = "";
			/*
			 * readLine is a bit quirky :
			 * it returns the content of a line MINUS the newline.
			 * it returns null only for the END of the stream.
			 * it returns an empty String if two newlines appear in a row.
			 */	
			while (( line = input.readLine()) != null){
				Matcher m = p.matcher(line);
				if (m.find()) {
					if (!sentence.toString().equals("")){
						String sent = sentence.toString();
						sentences.put(Integer.parseInt(sentID), sent);
					}
					sentID = m.group(1);
					continue;
				}
				if (line.startsWith("(")){
					sentence = new StringBuffer();
					sentence.append(line);
				}
				else{
					sentence.append(line.replaceAll("^(\t| )+", ""));	
				}
			}	
			if (!sentence.equals("")){
				String sent = sentence.toString();
				sentences.put(Integer.parseInt(sentID), sent);
				//System.out.println(sentence);
			}	
		}
		catch (FileNotFoundException e) {
			System.err.println(e.getMessage());
			//e.printStackTrace();
		}
		catch (IOException ex){
			ex.printStackTrace();
		}
		finally {
			try {
				if (input!= null) {
					input.close();
				}
			}
			catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}
	
	/*
	 * fileNegra is in western ISO encoding!
	 */
	public void enrichTrees(File fileNegra) {
		//Utils.printMemory("TigerCorpus - EnrichTrees - Start");
		InputStreamReader fr ;
		BufferedReader input = null;
		try {	
			fr = new InputStreamReader(new FileInputStream(fileNegra), Charset.forName("ISO-8859-1"));
			input = new BufferedReader(fr);
			//Utils.printMemory("TigerCorpus - EnrichTrees - 1");
			
			String lineOriginal = null;
			boolean inSent = false;
			int ctWord = -1;
			TigerTree currTree = null;
			Integer currSentID = null;
			Pattern pBOS = Pattern.compile("^#BOS\\s+([0-9]+)");
			Pattern pWhite = Pattern.compile("\\s+");
			Pattern pNode = Pattern.compile("^#[0-9]+");
			Pattern pMorph = Pattern.compile("\\.");
			Matcher m = null;
			Matcher m2 = null;
			String[] col = null;
			//Utils.printMemory("TigerCorpus - EnrichTrees - 2");
			/*
			 * readLine is a bit quirky :
			 * it returns the content of a line MINUS the newline.
			 * it returns null only for the END of the stream.
			 * it returns an empty String if two newlines appear in a row.
			 */	
			String line = null;
			while (( line = input.readLine()) != null) {
				// comment
				//if (line.startsWith("%%")) {
				if (line.charAt(0) == '%' && line.charAt(1) == '%') {
					continue;
				}
				// start of a sentence
				m = pBOS.matcher(line);
				if (m.find()) {
					inSent = true;
					currSentID = Integer.parseInt(m.group(1));
					currTree = trees.get(currSentID);
					ctWord = -1;
					//System.out.println("Enrich " + currSentID);
					if (debug && currTree != null && currSentID % 1000 == 0) {
						Utils.printMemory("TigerCorpus - Enrich trees " + currSentID);
					}
					//if (currSentID == 57) {
					//	System.out.println("test");
					//}
					continue;
				}
				// end of a sentence
				//if (line.startsWith("#EOS")) {
				if (line.charAt(0) == '#' && line.charAt(1) == 'E'
						&& line.charAt(2) == 'O') {
					inSent = false;
					continue;
				}
				// sentence data
				if (inSent && currTree != null) {
					ctWord++;
					//col = line.split("\\s+");
					lineOriginal = new String(line);
					col = pWhite.split(lineOriginal);
					if (col.length >= 6) { 
						boolean isLeaf = false;
						// word lemma tag morph edge parent (secedge secparent)*
						m2 = pNode.matcher(col[0]);
						//if (!col[0].matches("#[0-9]+")) { // lexical entry
						if (! m2.find()) {
							isLeaf = true;
							currTree.putLemma(ctWord, col[1]);
////							((LeafNode)currTree.getID2Node().get(ctWord)).setLemma(col[1]);
							currTree.putMorphology(ctWord, col[3], pMorph);
						}
					
						// secondary edges
						int s = 6;
						while (s+1 < col.length) {
							String label = col[s];
							if (label.startsWith("%%")) break;
							int endID = Integer.parseInt(col[s+1]);
							int startID;
							if (isLeaf) {
								startID = ctWord;
							}
							else {
								startID = Integer.parseInt(col[0].substring(1));
							}
							currTree.putSecEdge(label, startID, endID);
							s += 2;
						}
					}
				}
			}
			//Utils.printMemory("TigerCorpus - EnrichTrees - 3");
		} 
		catch (FileNotFoundException e) {
			System.err.println(e.getMessage());
			// e.printStackTrace();
		} 
		catch (IOException ex) {
			ex.printStackTrace();
		} 
		finally {
			try {
				if (input != null) {
					input.close();
				}
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}

	public void printTrees() {
		List<Integer> sortedList = new ArrayList<Integer>();
		for (Integer s : trees.keySet()) {
			sortedList.add(s);
		}
		Collections.sort(sortedList);
		for (Integer id : sortedList) {
			System.out.println("-- " + id + " --");
			System.out.println(trees.get(id));
		}
		
	}

}
