package pltag.german;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.regex.Pattern;



public class TigerTree {
	
	private static final String NKHD = "NKHD"; // new functional category for the head of an NP
	private static final String PCC = "PCC"; // comparative complement phrase, new phrasal category
	
	private String sourceString;
	private TreeNode root;
	private ArrayList<LeafNode> leafList;
	private HashMap<Integer, TreeNode> ID2Node;
	private HashMap<String, Trace> traceList;
	private ArrayList<SecondaryEdge> secondaryEdgesList;
	
	private String currentString; // needed to parse the bracketing into a tree
	private IDGenerator idgen = new IDGenerator(); // for the nodes that are inserted and that do not have a Tiger ID
	
	private SiblingStatus siblingStatus;
	private HashMap<Integer, TreeNode> foot2Root;
	private HashMap<Integer, TreeNode> root2Foot;

	public TigerTree(String sentence) {
		sourceString = sentence;
		currentString = sentence;
		leafList = new ArrayList<LeafNode>();
		ID2Node = new HashMap<Integer, TreeNode>();
		traceList = new HashMap<String, Trace>();
		secondaryEdgesList = new ArrayList<SecondaryEdge>();
		root = parseCurrentString();
		putNodesOnList(root);
		//System.out.println(traceList);
		//System.out.println(ID2Node);
		
		siblingStatus = new SiblingStatus();
		foot2Root = new HashMap<Integer, TreeNode>();
		root2Foot = new HashMap<Integer, TreeNode>();
	}
	
	
	public String getSourceString() {
		return sourceString;
	}
	
	public IDGenerator getIDGenerator() {
		return idgen;
	}
	
	public ArrayList<LeafNode> getLeafNodes() {
		return leafList;
	}
	
	public HashMap<String, Trace> getTraceList() {
		return traceList;
	}
	
	public HashMap<Integer, TreeNode> getFoot2Root() {
		return foot2Root;
	}
	
	public HashMap<Integer, TreeNode> getRoot2Foot() {
		return root2Foot;
	}

	private void putNodesOnList(TreeNode node) {
		// tree traversal to put the nodes onto the HashMap ID2Node
		if (node != null && node.getID() != -999) {
			ID2Node.put(node.getID(), node);
		}
		for (TreeNode child : node.getChildren()) {
			putNodesOnList(child);
		}
	}

	private TreeNode parseCurrentString() {
		String s = currentString.trim();
		 
		String id = "";
		int index = 1; // position after opening bracket
		boolean notEnd = true;
		String category;
		TreeNode currentNode;
		
		if (s.charAt(0) != '(') {
			return null;
		}
		
		while (s.charAt(index) != '(' & notEnd){
			//as long as no subcategory started
		
			if (s.charAt(index) == ')') { // leaf
				notEnd = false;
				String leaf;
				String catleaf = s.substring(1, index).trim();
				
				if (catleaf.startsWith("*T")) { // trace
					if (catleaf.contains("|||")) { 
						// nodeID is with the trace (antecedent is an inner node)
						category = catleaf.substring(0, catleaf.indexOf("|||"));
						id = catleaf.substring(catleaf.indexOf("|||")+3, catleaf.indexOf(" "));
						if (id.equals("0")) { // this cannot really happen, I think
							id = String.valueOf(idgen.getNewID());
						}
						else {
							 id = "-" + id;
						}
						leaf = catleaf.substring(catleaf.indexOf(" ")+1);
						currentNode = new LeafNode(Integer.parseInt(id), category, leaf, traceList);
					}
					else {
						// nodeID is with the antecedent (antecedent is a leaf node)
						category = catleaf.substring(0, catleaf.indexOf(" "));
						leaf = catleaf.substring(catleaf.indexOf(" ")+1);
						if (id.length() == 0) { // should always be the case
							id = String.valueOf(idgen.getNewID());
						}
						currentNode = new LeafNode(Integer.parseInt(id), category, leaf, traceList);
					}
				}
				else { // regular leaf node
					category = catleaf.substring(0, catleaf.indexOf(" "));	
					leaf = catleaf.substring(catleaf.indexOf(" ")+1, catleaf.indexOf("|||"));
					id = catleaf.substring(catleaf.indexOf("|||")+3);
					currentNode = new LeafNode(Integer.parseInt(id), category, leaf, traceList);
				}
				currentString = s.substring(index+1);
//				((LeafNode) currentNode).setLeafNumber(leaflist.size());//size should be the same as the position of the new entry.
				leafList.add((LeafNode)currentNode);
				return currentNode;
			}
			else{
				index++;
			}
		}
		
		// determine category and ID
		if (index == 1) { // there was not category	
			category = "VNODE";
			id = "999";
		}
		else {
			category = (s.substring(1, index)).trim();
			String[] ss = category.split("\\|\\|\\|");
			if (ss.length == 2) {
				category = ss[0];
				id = ss[1];
			}
		}
		if (id.length() == 0) {
			id = String.valueOf(idgen.getNewID());
		}
		currentNode = new InnerNode(Integer.parseInt(id), category, traceList);
		
		String sRest = s.substring(index).trim();
		currentString = sRest.trim();

		TreeNode lastChild = null;
		while (! currentString.startsWith(")")){
			TreeNode child = parseCurrentString();
			child.setParent(currentNode);
			currentNode.addChild(child);
			if (lastChild != null){
				child.setLeftSibling(lastChild);
				lastChild.setRightSibling(child);
			}
			lastChild = child;
			//currentString = currentString.trim();
		}
		if (currentString.startsWith(")")){
			currentString = currentString.substring(1);
		}

		return currentNode;
	}
	
	public HashMap<Integer, TreeNode> getID2Node() {
		return ID2Node;
	}
	
	public void removeNode(TreeNode node) {
		Integer id = node.getID();
		ID2Node.remove(id);
	}
	
	public TreeNode getRoot() {
		return root;
	}
	
	public void setRoot(TreeNode node) {
		root = node;
	}
	
	public void putLemma(Integer id, String lemma) {
		TreeNode node = ID2Node.get(id);
		if (node != null && node.isLeaf()) {
			((LeafNode) node).setLemma(lemma); 
		}
	}
	
	public void putMorphology(Integer id, String morph, Pattern p) {
		TreeNode node = ID2Node.get(id);
		if (node != null && node.isLeaf()) {
			((LeafNode) node).setMorphology(morph, p); 
		}
	}
	
	public void putSecEdge(String label, int startID, int endID) {
		TreeNode startNode = ID2Node.get(startID);
		TreeNode endNode = ID2Node.get(endID);
		SecondaryEdge secEdge = new SecondaryEdge(label, startNode, endNode);
		startNode.addSecondaryEdge(secEdge);
		endNode.addSecondaryEdge(secEdge);
		secondaryEdgesList.add(secEdge);
	}
	
	public ArrayList<SecondaryEdge> getSecondaryEdges() {
		return secondaryEdgesList;
	}

	public int countNodes() {
		return countNodes(root);
	}
	
	private int countNodes(TreeNode node) {
		int count = 1;
		for (TreeNode child : node.getChildren()) {
			count += countNodes(child);
		}
		return count;
	}
	
	public boolean isBinary() {
		return isBinary(root);
	}
	
	private boolean isBinary(TreeNode node) {
		if (node.getChildren().size() > 2) {
			return false;
		}
		for (TreeNode child : node.getChildren()) {
			boolean res = isBinary(child);
			if (!res) return false;
		}
		return true;
	}
	
	public String toString() {
		StringBuffer s = new StringBuffer();
		toString(s, root, 0);
		return s.toString();
	}
	
	private void toString(StringBuffer s, TreeNode node, int level) {
		for (int i = 0; i < level; i++) {
			s.append("  ");
		}
		s.append(node.toString());
		if (node.getNodeStatusBottom() == NodeStatus.PREDADJ) {
			s.append("   PREDADJ");
		}
		s.append("\n");
		for (TreeNode child : node.getChildren()) {
			toString(s, child, level+1);
		}
	}

	/*
	 * Groups nominal material governed by a PP under an inserted NP node.
	 */
	public void insertNPUnderPP(TreeNode node) {
		if (node.getPhrasalCategory().equals("PP")) {
			ArrayList<TreeNode> children = node.getChildren();
			// Find the daughters of PP which should be grouped.
			// At least of one them should have the functional label NK.
			int start = 0;
			int end = 0;
			boolean seenNK = false;
			for (int i = 0; i < children.size(); i++) {
				TreeNode child = children.get(i);
				// AC marks the "head(s)" of the PP
				if (child.getFunctionalCategory().equals("AC")) {
					if (child.getPhrasalCategory().matches("APPR|APPRART|CAC")) { // prepositions
						start = i+1;
						end = children.size();
					}
					if (child.getPhrasalCategory().matches("APPO|APZR")) { // postpositions
						end = i;
					}
				}
				if (!seenNK && child.getFunctionalCategory().equals("NK")) {
					seenNK = true;
				}
			}
			
			if (start < end && seenNK) {
				if (end-start == 1 && // exactly one daughter to group
					(children.get(start).getPhrasalCategory().matches("CNP|NP") || 
							children.get(start).getPhrasalCategory().matches("ADV|AVP|CAVP|CAP|AP|ADJD|ADJA|PROAV"))) {
					// we don't do anything (only change the functional category)
					// * sometimes there might already be an NP, especially if it is a coordination
					// * prepositions can also subcategorize for adverbs/adjectives (which we do not want 
					//   to group under NP)
					children.get(start).setFunctionalCategory("PC");
				}
				else {
					// insert a new NP node with functional label PC (prepositional complement)
					TreeNode np = new InnerNode(idgen, "NP-PC");
					ID2Node.put(np.getID(), np);
					for (int i = 0; i < end-start; i++) {
						TreeNode child = node.removeChild(start);
						np.addChild(child);
						child.setParent(np);
					}
					node.addChild(start, np);
					np.setParent(node);
				}
			}
			
		}
		// recursion
		for (TreeNode child : node.getChildren()) {
			insertNPUnderPP(child);
		}
		
	}

	/*
	 * Introduces the maximal projections AVP, AP, VP and DP
	 */
	public void introduceMaxProjections(TreeNode node) {		
		String cat = node.getPhrasalCategory();
		String function = node.getFunctionalCategory();
		// AVP
		if (cat.matches("ADV|PTKNEG") && !function.matches("HD|PH|AVC|UC")) {
			addUnaryNodeAbove(node, "AVP", function, "HD");
		}
		
		// AP
		else if (cat.matches("ADJA|ADJD|MTA|AA") && !function.matches("HD|PH|ADC")) {
			addUnaryNodeAbove(node, "AP", function, "HD");
		}
		else if (cat.matches("CARD|NM")) {
			if (function.matches("CJ") && node.getParent().getPhrasalCategory().equals("CAP")) {
				addUnaryNodeAbove(node, "AP", function, "HD");
			}
		}
		
		// VP
		else if ((cat.matches("VMPP|VAPP|VVPP|VAINF|VVINF") && !function.matches("HD")) ||
				(cat.matches("VVIZU|VZ|CVZ") && !function.matches("HD|CJ"))) {
			addUnaryNodeAbove(node, "VP", function, "HD");
		}
		
		// VZ/VP in coordination
		else if (cat.matches("VVIZU") && function.matches("CJ")) {
			if (node.getParent().getPhrasalCategory().equals("CVZ")) {
				addUnaryNodeAbove(node, "VZ", function, "HD");
			}
			else if (node.getParent().getPhrasalCategory().equals("CVP")) {
				addUnaryNodeAbove(node, "VP", function, "HD");
			}
		}
		else if (cat.matches("VZ") && function.matches("CJ")) {
			if (!node.getParent().getPhrasalCategory().equals("CVZ")) {
				addUnaryNodeAbove(node, "VP", function, "HD");
			}
		}
		
		// DP
		else if (cat.matches("ART|PDAT|PPOSAT|PRELAT|PWAT|PIAT") ||
				(cat.matches("PDS") && function.matches("AG"))) {
			if (!function.matches("HD")) {
				addUnaryNodeAbove(node, "DP", function, "HD");
			}
			else if (function.matches("HD")) {
				// in Tiger determiners project to AP, let's change that
				node.getParent().setPhrasalCategory("DP");
			}
		
		}		
		
		// AC
		// TODO: introduce AC everywhere!?!
		else if (cat.matches("APPR|APPRART|APPO|APZR")) {
			if (function.matches("CJ") && node.getParent().getPhrasalCategory().matches("CAC")) {
				addUnaryNodeAbove(node, "AC", function, "HD");
			}
			else if (function.matches("HD")) {
				node.getParent().setPhrasalCategory("AC");
			}
		}
		
		//recursion
		for (int i = 0; i < node.getChildren().size(); i++) {
			TreeNode child = node.getChildren().get(i);
			introduceMaxProjections(child);
		}	
	}
	
	public void introduceNPs(TreeNode node) {
		disambiguateNPs(node);
		introduceNPProjections(node);
	}
	
	private void introduceNPProjections(TreeNode node) {
		String cat = node.getPhrasalCategory();
		String function = node.getFunctionalCategory();
		if ((cat.matches("NN|NE|PIS|PPER|PPOSS|PRELS|PRF|PWS|PN") ||
				(cat.matches("PDS") && (!function.matches("AG")))) &&
			!function.matches("PH|NKHD|PNC|NMC|HD|NK|ADC")) {
			addUnaryNodeAbove(node, "NP", function, "NKHD");
		}
		if (cat.matches("CARD|NM") && function.equals("CJ") &&
				node.getParent().getPhrasalCategory().equals("CNP")) {
			addUnaryNodeAbove(node, "NP", function, "NKHD");
		}
		for (int i = 0; i < node.getChildren().size(); i++) {
			TreeNode child = node.getChildren().get(i);
			introduceNPProjections(child);
		}
	}

	private void disambiguateNPs(TreeNode node) {
		String cat = node.getPhrasalCategory();
		if (cat.matches("NP")) {
			ArrayList<TreeNode> children = node.getChildren();
			ArrayList<TreeNode> potHeads = new ArrayList<TreeNode>(5); // potential heads
			int indDP = -1; // index of DP
			
			// go through the children and look for potential heads
			for (int i = 0; i < children.size(); i++) {
				TreeNode child = children.get(i);
				String childCat = child.getPhrasalCategory();
				String childFunction = child.getFunctionalCategory();
				if (childFunction.equals("PH")) {
					// this is the head, no further disambiguation needed
					break;
				}
				else if (childFunction.equals("NK")) {
					if (!childCat.matches("DP|AP|AVP|CAP|AC")) {
						potHeads.add(child);
					}
					else if (childCat.equals("DP")) {
						indDP = i;
					}
				}
			}
			if (potHeads.size() == 1) {
				// the head is clear
				potHeads.get(0).setFunctionalCategory(NKHD);
			}
			else if (potHeads.size() == 2) {
				String potHead1Cat = potHeads.get(1).getPhrasalCategory();
				if (potHeads.get(0).getPhrasalCategory().matches("NN|NP") &&
						potHead1Cat.matches("NN|NP")) {
					// eine Tracht Prügel, der Landeswährung Lira, eine Art [Füllhorn des Wissens],
					// über den Problemkomplex [multikulturelle Gesellschaft], in 3 Metern Abstand,
					// bis Ende März, Anfang Juni
					// -> left component is the head
					potHeads.get(0).setFunctionalCategory(NKHD);
					// insert NP for non-head
					if (potHead1Cat.equals("NN")) {
						addUnaryNodeAbove(potHeads.get(1), "NP", "NK", "NKHD");
					}
				}
				else if (potHeads.get(1).getPhrasalCategory().matches("NE|PN")) {
					TreeNode mother = node.getParent();
					int i = mother.getChildren().indexOf(node) - 1; // left sibling
					if (mother.getPhrasalCategory().equals("PP") && i > -1 &&
							mother.getChildren().get(i).getPhrasalCategory().equals("APPRART")) {
						// im Land Mosambik
						// -> left component is the head
						potHeads.get(0).setFunctionalCategory(NKHD);
						// insert NP for non-head
						addUnaryNodeAbove(potHeads.get(1), "NP", "NK", "NKHD");
					}
					else if (indDP > -1) {
						// des Wirtschaftsmagazines Fortune
						// -> left component is the head
						potHeads.get(0).setFunctionalCategory(NKHD);
						// insert NP for non-head
						addUnaryNodeAbove(potHeads.get(1), "NP", "NK", "NKHD");
					}
					else {
						// Bundeskanzler Helmut Kohls (Abtritt)
						// -> right component is the head
						potHeads.get(1).setFunctionalCategory(NKHD);
						// insert NP for non-head
						if (!potHeads.get(0).getPhrasalCategory().matches("NP")) {
							addUnaryNodeAbove(potHeads.get(0), "NP", "NK", "NKHD");
						}
					}
					// TODO: disambiguate NE+NN
				}
				// TODO: more fine-grained disambiguation could also be done for NPs
				// with 3 or more potential heads.
			}
			if (potHeads.size() == 0) { // only DPs, APs etc. or annotation error -> the rightmost is the head
				for (int j = children.size()-1; j >= 0; j--) {
					TreeNode child = children.get(j);
					if (child.getFunctionalCategory().matches("NK")) {
						child.setFunctionalCategory(NKHD);
						break;
					}
				}
			}
		}
		//recursion
		for (int i = 0; i < node.getChildren().size(); i++) {
			TreeNode child = node.getChildren().get(i);
			disambiguateNPs(child);
		}
	}

	private TreeNode addUnaryNodeAbove(TreeNode node, String phrasalCat, String funcCat, String funcCatChild) {
		TreeNode newNode = new InnerNode(idgen, phrasalCat + "-" + funcCat);
		ID2Node.put(newNode.getID(), newNode);
		TreeNode mother = node.getParent();
		int ind = mother.getChildren().indexOf(node);
		mother.removeChild(ind);
		mother.addChild(ind, newNode);
		newNode.setParent(mother);
		newNode.addChild(node);
		node.setParent(newNode);
		
		// make the single child of the new node its head
		node.setFunctionalCategory(funcCatChild);
		
		// if the child was a filler, the new node should have the filler status now
		// (but we don't change the Trace object for now)
		if (node.isFiller()) {
			node.setFiller(false);
			newNode.setFiller(true);
			newNode.trace = node.trace;
			newNode.traceID = node.traceID;
		}
		
		// if the child was start or end point of a secondary edge, the
		// new node should have this status now.
		for (int i = 0; i < node.getSecondaryEdges().size(); i++) {
			SecondaryEdge sec = node.getSecondaryEdges().get(i);
			if (sec.getStartNode() == node) {
				sec.setStartNode(newNode);
			}
			else if (sec.getEndNode() == node) {
				sec.setEndNode(newNode);
			}
			node.getSecondaryEdges().remove(sec);
			newNode.addSecondaryEdge(sec);
		}
		
		
		return newNode;
	}
	
	private TreeNode addEmptyChild(TreeNode mother, int i, String leaf, String phrasalCat, String funcCat) {
		TreeNode newNode = new LeafNode(idgen, phrasalCat + "-" + funcCat, leaf);
		ID2Node.put(newNode.getID(), newNode);
		if (i < mother.getChildren().size()) {
			mother.addChild(i, newNode);
		}
		else {
			mother.addChild(newNode);
		}
		newNode.setParent(mother);
		// empty element, but without filler, i.e. not on the traceList
		((LeafNode)newNode).setIsTrace();
		// Fix leafList
		LeafNode nextLeaf = findNextLeaf(mother, i+1);
		if (nextLeaf != null) {
			int ind = leafList.indexOf(nextLeaf);
			leafList.add(ind, (LeafNode)newNode);
		}
		else {
			leafList.add((LeafNode)newNode);
		}
		
		return newNode;
	}
	
	private LeafNode findNextLeaf(TreeNode mother, int ind) {
		if (ind < mother.getChildren().size()) {
			TreeNode node = mother.getChildren().get(ind);
			while (!node.isLeaf()) {
				node = node.getChildren().get(0);
			}
			return (LeafNode)node;
		}
		else {
			TreeNode grandMother = mother.getParent();
			if (grandMother == null) {
				return null;
			}
			int i = grandMother.getChildren().indexOf(mother);
			return findNextLeaf(grandMother, i+1);
		}
	}


	public void identifyMatrixClauses(TreeNode node) {
		
		//recursion
		for (int i = 0; i < node.getChildren().size(); i++) {
			TreeNode child = node.getChildren().get(i);
			identifyMatrixClauses(child);
		}
		
		if (node.getPhrasalCategory().matches("VV.*")) {
			// full verb which has a sibling with function OC (clausal object)
			TreeNode parent = node.getParent();
			//TreeNode grandParent = parent.getParent();
			for (int i = 0; i < parent.getChildren().size(); i++) {
				TreeNode child = parent.getChildren().get(i);
				if (!child.isHead() && child.getFunctionalCategory().matches("OC")) {
					// Find the future root of the auxiliary tree and mark it ADJ
					while (!parent.getPhrasalCategory().equals("S")) {
						if (parent.getPhrasalCategory().equals("VP")) {
							parent = parent.getParent();
						}
						else {
							System.err.println("### Error identifying matrix clauses, no VP: ");
							System.err.println(Utils.subtreeToString(parent));
							break;
						}
					}
					if (parent.getPhrasalCategory().equals("S")) { 
						parent.setNodeStatusBottom(NodeStatus.PREDADJ);
					}
					else {
						System.err.println("### Error identifying matrix clauses, no S: ");
						System.err.println(Utils.subtreeToString(parent));
					}
//					if (parent.getPhrasalCategory().equals("S")) { // the full verb is finite
//						parent.setNodeStatusBottom(NodeStatus.PREDADJ);
//					}
//					else if (parent.getPhrasalCategory().equals("VP") && // or the full verb is infinite
//							grandParent != null && grandParent.getPhrasalCategory().equals("S")) {
//						grandParent.setNodeStatusBottom(NodeStatus.PREDADJ);
//					}
//					else {
//						System.err.println("### Error identifying matrix clauses: ");
//						System.err.println(Utils.subtreeToString(parent));
//					}
					// This child will be the future foot node.
					// Check whether an additional S node needs to be inserted.
					if (child.getPhrasalCategory().matches("VP")) { 
						TreeNode newNode = addUnaryNodeAbove(child, "S", "OC", "HD");
						// fix the heads (they have already been set before calling this method)
						newNode.setHeadChild(child);
						child.setHead();
						// add empty subject
						addEmptyChild(newNode, 0, "*0*", "E", "SB");
						newNode.setNodeStatus(NodeStatus.ARG);
					}
					// For coordinations some more work has to be done:
					// An additional S node is inserted abouve each conjunct.
					else if (child.getPhrasalCategory().matches("CVP")) {
						child.setPhrasalCategory("CS");
						child.setNodeStatus(NodeStatus.ARG);
						for (int j = 0; j < child.getChildren().size(); j++) {
							TreeNode grandChild = child.getChildren().get(j);
							if (grandChild.getFunctionalCategory().equals("CJ") && 
									grandChild.getPhrasalCategory().matches("VP")) {
								TreeNode newNode = addUnaryNodeAbove(grandChild, "S", "CJ", "HD");
								// fix the heads (they have already been set before calling this method)
								newNode.setHeadChild(grandChild);
								grandChild.setHead();
								// add empty subject
								addEmptyChild(newNode, 0, "*0*", "E", "SB");
							}
						}
					}
					else {
						child.setNodeStatus(NodeStatus.ARG);
					}
					break;
				}
			}
		}

	}

	public void reattachPunctuation() {
		// Try to attach quotes to the same node: `` and ''
		//findMatchingPairOfQuotes(0);
		// TODO
		
		// try to attach sentence-final punctuation lower,
		// such that we can get rid of VNODE later
//		if (root.getChildren().size() == 2 &&
//				Utils.isPunct(root.getChildren().get(1).getPhrasalCategory()) &&
//				!root.getChildren().get(0).isLeaf()) {
//			TreeNode punct = root.getChildren().get(1);
//			TreeNode newRoot = root.getChildren().get(0);
//			TreeNode oldRoot = root;
//			newRoot.addChild(punct);
//			punct.setParent(newRoot);
//			oldRoot.removeChild(punct);
//		}
		
		ArrayList<TreeNode> partition1 = new ArrayList<TreeNode>();
		ArrayList<TreeNode> partition2 = new ArrayList<TreeNode>();
		TreeNode newRoot = null;
		
		for (TreeNode child : root.getChildren()) {
			if (Utils.isPunct(child.getPhrasalCategory())) {
				if (newRoot == null) {
					partition1.add(child);
				}
				else {
					partition2.add(child);
				}
			}
			else {
				if (newRoot == null) {
					newRoot = child;
				}
				else {
					newRoot = null;
					break;
				}
			}
		}
		if (newRoot != null && !newRoot.isLeaf()) {
			for (int i = partition1.size()-1; i >= 0; i--) {
				TreeNode punct = partition1.get(i);
				newRoot.addChild(0, punct);
				punct.setParent(newRoot);
				root.removeChild(punct);
			}
			for (int i = 0; i < partition2.size(); i++) {
				TreeNode punct = partition2.get(i);
				newRoot.addChild(punct);
				punct.setParent(newRoot);
				root.removeChild(punct);
			}
		}
	}
	
	public void removeRoot() {
		if (root.getChildren().size() == 1) {
			TreeNode newRoot = root.getChildren().get(0);
			TreeNode oldRoot = root;
			root = newRoot;
			removeNode(oldRoot);
			getRoot().setNodeStatus(NodeStatus.ARG);
			newRoot.setParent(null); // TODO ok??
		}
		
	}

	private void findMatchingPairOfQuotes(int start) {
		int opening = -1;
		for (int i = start; i < leafList.size(); i++) {
			LeafNode leaf = (LeafNode)leafList.get(i);
			if (leaf.getLeafString().equals("``")) {
				if (opening == -1) {
					opening = i;
				}
				else {
					findMatchingPairOfQuotes(i);
				}
			}
			else if (leaf.getLeafString().equals("''")) {
				// TODO
			}
		}
		
	}

	public void introducePCC(TreeNode node) {
		
		//recursion
		for (int i = 0; i < node.getChildren().size(); i++) {
			TreeNode child = node.getChildren().get(i);
			introducePCC(child);
		}
		
		// look for comparative conjunctions (CM)
		if (node.getFunctionalCategory().equals("CM")) {
			// node of category XX
			TreeNode parent = node.getParent();
			if (parent.getPhrasalCategory().matches(PCC+"|S")) {
				// somehow we have already been here;
				// we don't insert PCC for S (in this case, CM works more like a 
				// complementizer)
				return;
			}
			if (node.isLeaf() && ((LeafNode)node).isTrace()) {
				return;
			}
			ArrayList<TreeNode> nodesToHangHigher = new ArrayList<TreeNode>();
			boolean everythingIsOk = true;
			for (TreeNode child : parent.getChildren()) {
				// modifiers to the left of CM are ok, we will also
				// hang them higher. if there is something else there,
				// we do not introduce PCC (those seem to be annotation errors)
				if (child.getFunctionalCategory().matches("MO|NG")) {
					nodesToHangHigher.add(child);
				}
				else if (child.getFunctionalCategory().equals("CM")) {
					nodesToHangHigher.add(child);
					break;
				}
				else {
					everythingIsOk = false;
					break;
				}
			}
			if (everythingIsOk && nodesToHangHigher.size() > 0) {
				// the new node, has the functional category of XX
				TreeNode newNode = new InnerNode(idgen, PCC + "-" + parent.getFunctionalCategory());
				ID2Node.put(newNode.getID(), newNode);
				TreeNode grandParent = parent.getParent();
				int ind = grandParent.getChildren().indexOf(parent);
				grandParent.removeChild(ind);
				grandParent.addChild(ind, newNode);
				newNode.setParent(grandParent);
				// hang the nodes to the new node
				for (TreeNode child : nodesToHangHigher) {
					parent.removeChild(child);
					newNode.addChild(child);
					child.setParent(newNode);
				}
				newNode.addChild(parent);
				parent.setParent(newNode);
								
				// new functional category for XX
				parent.setFunctionalCategory("COB");
				
				// if XX was a filler, the new node should have the filler status now
				// (but we don't change the Trace object for now)
				if (parent.isFiller()) {
					parent.setFiller(false);
					newNode.setFiller(true);
					newNode.trace = parent.trace;
					newNode.traceID = parent.traceID;
				}
				
				// if XX was the first CJ in a coordination, the newNode will be
				// the foot node later. accordingly we need to change the phrasal
				// category of the coordination node.
				if (Utils.isCoord(grandParent)) {
					for (TreeNode child : grandParent.getChildren()) {
						if (child.getFunctionalCategory().equals("CJ")) {
							if (child ==  newNode) {
								grandParent.setPhrasalCategory(PCC);
								grandParent.setCoordinatedCat();
							}
							break;
						}
					}
				}
			}
			
		}
	}
	
	
	public void reattachSomeFillers() {
		if (traceList.isEmpty()) {
			return;
		}
		while (reattachSomeFillers(root));
	}
		
	private boolean reattachSomeFillers(TreeNode node) {
		
		//recursion
		for (int i = 0; i < node.getChildren().size(); i++) {
			TreeNode child = node.getChildren().get(i);
			boolean reattached = reattachSomeFillers(child);
			if (reattached) {
				return reattached;
			}
		}
		
		boolean reattach = false;
		
		// verbal arguments that lead to crossing branches are sometimes minimally attached higher
		// in a way that makes them dependent of a modal or auxiliary 
		// (e.g. die Partei wird sich auf Dauer nicht stabilisieren können)
		// we attach them again one level higher in order to avoid unbound traces in the 
		// elementary trees later.
		if (node.isFiller()) {
			TreeNode parent = node.getParent();
			TreeNode grandParent = parent.getParent();
			if (node.getFunctionalCategory().matches(Utils.verbalArgumentFunctionalCats) &&
					parent != null &&
					grandParent != null &&
					parent.getPhrasalCategory().matches("VP") &&
					grandParent.getPhrasalCategory().matches("VP|S") &&
					node.trace.getHole().getParent().getPhrasalCategory().matches("VP|S") &&
					(parent.isLeftMostChild(node) || parent.isRightMostChild(node))) {
				for (int i = 0; i < parent.getChildren().size(); i++) {
					TreeNode sibling = parent.getChildren().get(i);
					if (sibling.getFunctionalCategory().matches("HD") &&
							Utils.isAuxiliaryOrModal(sibling)) {
						reattach = true;
						break;
					}
				}
				if (reattach) {
					int index = grandParent.indexOfChild(parent);
					if (parent.isLeftMostChild(node)) {
						grandParent.addChild(index, node);
					}
					else { // isRightMostChild
						grandParent.addChild(index+1, node);
					}
					node.setParent(grandParent);
					parent.removeChild(node);
				}
			}
		}
		
		return reattach;
	}
	
	public void fixCoordinations(TreeNode node) {
		// recursion
		for (TreeNode child : node.getChildren()) {
			fixCoordinations(child);
		}
				
		// coordination -> fix category name: remove the C
		// Replace CO by the phrasal category of its left-most CJ daughter
		// TODO: is this reversible?!
		if (node.getPhrasalCategory().matches(Utils.coordinatedCats)) {
			node.setPhrasalCategory(node.getPhrasalCategory().substring(1));
			node.setCoordinatedCat();
		
			if (node.getPhrasalCategory().equals("O")) {
				for (TreeNode child : node.getChildren()) {
					if (child.getFunctionalCategory().equals("CJ")) {
						String childCat = child.getPhrasalCategory();
						//if (childCat.matches(Utils.coordinatedCats)) {
						//	childCat = childCat.substring(1);
						//}
						node.setPhrasalCategory(childCat);
						node.setUnlikeCoordination(true);
						break;
					}
				}
			}
		}
	}
	
	public void determineSiblingStatus() {
		getRoot().setNodeStatus(NodeStatus.ARG);
		determineChildrenStatus(getRoot());
	}
	
	private void determineChildrenStatus(TreeNode node) {
		siblingStatus.determineStatusOfChildren(node);
		
		// recursion
		for (TreeNode child : node.getChildren()) {
			determineChildrenStatus(child);
		}
	}
	
	public void identifyAuxiliaryTrees(TreeNode node) {

		// check whether one of the children is the foot of a predicative auxiliary tree.
		// then we need to mark the root.
		// so far restricted to S-OC and VP-OC (auxiliary trees for infinite auxililaries and modals)
		CHILDLOOP : for (int i = 0; i < node.getChildren().size(); i++) {
			TreeNode child = node.getChildren().get(i);
			if (child.getNodeStatus() == NodeStatus.ARG && child.getFunctionalCategory().matches("OC") &&
					child.getPhrasalCategory().matches("S|VP") && !child.isFiller()) {
				
				// search upwards for the future root of the aux tree
				TreeNode potentialRoot = node;
				while (potentialRoot != null) {
						// the same phrasal categories -> found the future root
						// TODO: it is marked PREDADJ at the moment. should be ADJ later
						if (potentialRoot.getPhrasalCategory().equals(child.getPhrasalCategory()) 
								|| (potentialRoot.getPhrasalCategory().equals("S") && child.getPhrasalCategory().equals("VP"))) {
							
							if (potentialRoot.getPhrasalCategory().equals("VP")) {
								// for VP we only extract aux trees where the anchor is a modal or auxiliary at the moment
								if (potentialRoot.getHeadChild().getPhrasalCategory().matches("VZ")) {
									if (potentialRoot.getHeadChild().getHeadChild().getPhrasalCategory().matches("(VA|VM).*")) {
										potentialRoot.setNodeStatusBottom(NodeStatus.PREDADJ);
										foot2Root.put(child.getID(), potentialRoot);
										root2Foot.put(potentialRoot.getID(), child);
										break CHILDLOOP;
									}
								}
								else if (potentialRoot.getHeadChild().getPhrasalCategory().matches("(VA|VM).*")) {
									potentialRoot.setNodeStatusBottom(NodeStatus.PREDADJ);
									foot2Root.put(child.getID(), potentialRoot);
									root2Foot.put(potentialRoot.getID(), child);
									break CHILDLOOP;
								}
							}
							else if (potentialRoot.getPhrasalCategory().equals("S")) {
							
								if (child.getPhrasalCategory().equals("VP")) {
									// introduce an additional S node to make the recursion possible
									
									TreeNode newNode = addUnaryNodeAbove(child, "S", "OC", "HD");
									// fix the heads (they have already been set before calling this method)
									newNode.setHeadChild(child);
									child.setHead();
										// add empty subject
										//addEmptyChild(newNode, 0, "*0*", "E", "SB");
									newNode.setNodeStatus(NodeStatus.ARG);
									child = newNode;
								}
								
								potentialRoot.setNodeStatusBottom(NodeStatus.PREDADJ);
								foot2Root.put(child.getID(), potentialRoot);
								root2Foot.put(potentialRoot.getID(), child);
								
								// TAG can handle cases of scrambling where one daughter of the embedded clause
								// is moved to the matrix clause.
								// However, the flat Tiger annotation does not allow for it.
								// So if we see such a case, we re-attach the filler to a newly introduced level
								// to make such an analysis possible.
								TreeNode leftmostDaughter = potentialRoot.getChildren().get(0); 
								if (leftmostDaughter.isFiller() &&
										leftmostDaughter.getFunctionalCategory().matches(Utils.verbalArgumentFunctionalCats)) {
									// leftmost daughter of pot. root is an argument filler
									TreeNode hole = traceList.get(leftmostDaughter.traceID).getHole();
									// check whether the corresponding trace (hole) is contained within the subtree
									// rooted by the foot. Otherwise this transformation would not
									// achieve our target.
									if (child.subtreeContains(hole)) {
										// new S node;
										// the old S node (potential root) becomes its head
										TreeNode newS = addUnaryNodeAbove(potentialRoot, "S", 
														potentialRoot.getFunctionalCategory(), "HD");
										newS.setNodeStatus(potentialRoot.getNodeStatus());
										if (potentialRoot.isHead()) {
											newS.setHead();
											if (newS.getParent() != null) {
												newS.getParent().setHeadChild(newS);
											}
										}
										if (potentialRoot == root) {
											root = newS;
										}
										newS.setHeadChild(potentialRoot);
										potentialRoot.setHead();
										// attach filler to new S node
										potentialRoot.removeChild(leftmostDaughter);
										newS.addChild(0, leftmostDaughter);
										leftmostDaughter.setParent(newS);
										
										if (foot2Root.containsKey(potentialRoot.getID())) {
											// the potential root of the current auxiliary tree,
											// already serves as the foot node of some other auxiliary tree.
											// the new S node has to be this foot node now.
											TreeNode otherRoot = foot2Root.get(potentialRoot.getID());
											foot2Root.remove(potentialRoot.getID());
											foot2Root.put(newS.getID(), otherRoot);
											root2Foot.put(otherRoot.getID(), newS);
										}
									}
								}
								
								break CHILDLOOP;
							}
						}
						
						if (potentialRoot.isHead()){
							// follow the head trunk,
							// skip auxiliary trees that have already been found
							if (foot2Root.containsKey(potentialRoot.getID())) {
								potentialRoot = foot2Root.get(potentialRoot.getID());
							}
							potentialRoot = potentialRoot.getParent();
						}
						else {
							break;
						}
				}
			}
		}
		
		// recursion
		for (int i = 0; i < node.getChildren().size(); i++) {
			TreeNode child = node.getChildren().get(i);
			identifyAuxiliaryTrees(child);
		}
	}
	

	public boolean dominates(TreeNode node1, TreeNode node2) {
		if (node1 == null || node2 == null) {
			return false;
		}
		for (TreeNode child : node1.getChildren()) {
			if (child == node2) {
				return true;
			}
			else {
				if (dominates(child, node2)) {
					return true;
				}
			}
		}
		return false;
	}

	public void deleteModifierTraces(TigerTree treeCopy) {
		
		Iterator<Map.Entry<String,Trace>> iter = traceList.entrySet().iterator();
		while (iter.hasNext()) {
		    Map.Entry<String,Trace> entry = iter.next();
		    TreeNode trace = entry.getValue().getHole();
		    if (trace.getNodeStatus() == NodeStatus.MOD) {
				TreeNode pa = trace.getParent();
				pa.removeChild(trace);
				pa.setHadTraceChild(true);
				ID2Node.remove(trace.getID());
				leafList.remove(trace);
				iter.remove();
				
				// in copy
				String id = entry.getKey();
				TreeNode trace2 = treeCopy.getTraceList().get(id).getHole();
				TreeNode pa2 = trace2.getParent();
				pa2.removeChild(trace2);
				pa2.setHadTraceChild(true);
				treeCopy.getID2Node().remove(trace2.getID());
				treeCopy.getLeafNodes().remove(trace2);
				treeCopy.getTraceList().remove(id);
				
			}
		}
		
//		for (Trace t : traceList.values()) {
//			TreeNode trace = t.getHole();
//			if (trace.getNodeStatus() == NodeStatus.MOD) {
//				TreeNode pa = trace.getParent();
//				pa.removeChild(trace);
//				ID2Node.remove(trace.getID());
//				leafList.remove(trace);
//			}
//		}
		
	}

	public void deleteAllTraces(TigerTree treeCopy) {
		Iterator<Map.Entry<String,Trace>> iter = traceList.entrySet().iterator();
		while (iter.hasNext()) {
		    Map.Entry<String,Trace> entry = iter.next();
		    TreeNode trace = entry.getValue().getHole();
			TreeNode pa = trace.getParent();
			pa.removeChild(trace);
			pa.setHadTraceChild(true);
			ID2Node.remove(trace.getID());
			leafList.remove(trace);
			iter.remove(); 
			
			// in copy
			String id = entry.getKey();
			TreeNode trace2 = treeCopy.getTraceList().get(id).getHole();
			TreeNode pa2 = trace2.getParent();
			pa2.removeChild(trace2);
			pa2.setHadTraceChild(true);
			treeCopy.getID2Node().remove(trace2.getID());
			treeCopy.getLeafNodes().remove(trace2);
			treeCopy.getTraceList().remove(id);
		}
	}

	public void deleteTraces(ArrayList<String> tracesToDelete, TigerTree treeCopy,
				boolean regularAdjunction) {
		for (String traceID : tracesToDelete) {
			TreeNode trace = traceList.get(traceID).getHole();
			TreeNode pa = trace.getParent();
			pa.removeChild(trace);
			pa.setHadTraceChild(true);
			ID2Node.remove(trace.getID());
			leafList.remove(trace);
			traceList.remove(traceID);
			
			// in copy
			TreeNode trace2 = treeCopy.getTraceList().get(traceID).getHole();
			TreeNode pa2 = trace2.getParent();
			pa2.removeChild(trace2);
			pa2.setHadTraceChild(true);
			treeCopy.getID2Node().remove(trace2.getID());
			treeCopy.getLeafNodes().remove(trace2);
			treeCopy.getTraceList().remove(traceID);
			
			// also remove node that had been introduced for adjunction
			if (regularAdjunction) {
				if (pa.getChildren().size() == 1
						&& pa.getChildren().get(0).insertedForBinarization()
						&& pa.getChildren().get(0).getPhrasalCategory().equals(pa.getPhrasalCategory())) {
					TreeNode onlyChild = pa.getChildren().get(0);
					ArrayList<TreeNode> granChildren = onlyChild.getChildren();
					pa.removeChild(onlyChild);
					ID2Node.remove(onlyChild.getID());
					
					for (TreeNode child : granChildren) {
						pa.addChild(child);
						child.setParent(pa);
						if (child.isHead()) {
							pa.setHeadChild(child);
						}
					}
				}
			}
		}
		
	}



}
