package pltag.german.analysis;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import pltag.german.LexEntry;

public class Analysis {
	
	private File argTreesFile;
	private File modTreesFile;
	private File adjTreesFile;
	private File predTreesFile;
	private File ambiguityLexFile;
	private File ambiguityLemmaFile;
	private File templateFreqFile;
	private File sumFile;
	private File growthFile;
	private File growthCoreFile;
	private File matFile;
	private File lexiconFile;
	private Writer fileWriter;
	
	private boolean extendedCategories;
	private boolean morphology;
	private boolean secondaryEdgeMarker;
	private int grammarGrowthSteps; // data points (x)
	private int maxAnchors;
	
	private List<ElementaryTree> templatesTokFreqSorted =  new ArrayList<ElementaryTree>();
	
	private HashMap<String, ArrayList<String>> argTrees = new HashMap<String, ArrayList<String>>();
	private List<ElementaryTree> argTreesTokFreqSorted =  new ArrayList<ElementaryTree>();
	
	private HashMap<String, ArrayList<String>> modTrees = new HashMap<String, ArrayList<String>>();
	private List<ElementaryTree> modTreesTokFreqSorted =  new ArrayList<ElementaryTree>();
	
	private HashMap<String, ArrayList<String>> adjTrees = new HashMap<String, ArrayList<String>>();
	private List<ElementaryTree> adjTreesTokFreqSorted =  new ArrayList<ElementaryTree>();
	
	private HashMap<String, ArrayList<String>> predTrees = new HashMap<String, ArrayList<String>>();
	private List<ElementaryTree> predTreesTokFreqSorted =  new ArrayList<ElementaryTree>();
	
	private HashMap<String, ArrayList<String>> predTreesUncombined = new HashMap<String, ArrayList<String>>();
	
	private HashMap<String, ArrayList<String>> word2Tree = new HashMap<String, ArrayList<String>>();
	private List<WordFreqPair> word2TreeFreqSorted = new ArrayList<WordFreqPair>();
	private HashMap<String, ArrayList<String>> lemma2Tree = new HashMap<String, ArrayList<String>>();
	private List<WordFreqPair> lemma2TreeFreqSorted = new ArrayList<WordFreqPair>();
	
	private List<WordFreqPair> templateStorage = new ArrayList<WordFreqPair>(); // abusing WordFreqPair
	private HashSet<String> corePredLexicon = new HashSet<String>();
	private HashSet<String> coreCanonLexicon = new HashSet<String>();
	
	private HashMap<String, Integer> MAT = new HashMap<String, Integer>();
	
	private TreeMap<Integer, Integer> numTreesInPrecombinedPredTrees = new TreeMap<Integer, Integer>();
	
	private TreeMap<String, ArrayList<String>> lexicon = new TreeMap<String, ArrayList<String>>();
	
	private Pattern pTab = Pattern.compile("\\t");
	private Pattern pWhite = Pattern.compile("\\s+");
	private Pattern pAnchor = Pattern.compile("\\s[^\\s]+?<>");
	private Pattern pExtCat = Pattern.compile("-[^\\s\\*\\(\\)\\[]+?[\\^\\[\\$]");
	private Pattern pExtCat2 = Pattern.compile("-\\w{2,4}");
	private Pattern pMorph = Pattern.compile("\\[\\w+\\]");
	private Pattern pTrace = Pattern.compile("\\*T(\\d+)\\*");
	private Pattern pIndex = Pattern.compile("([_\\^])(\\d)");
	private Pattern pSecondaryEdgeMarker = Pattern.compile("\\$\\$.+?\\$\\$");
	private Pattern pCoAnchor = Pattern.compile("\\^\\d+_\\d+");
	private Pattern pTraceTree = Pattern.compile("\\(\\*T(\\d+)\\*[^\\)]+ -<>\\)");
	private Pattern pFiller = Pattern.compile("-\\*T\\d+\\*");
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		boolean extendedCategories = false;
		boolean morphology = false;
		boolean secondaryEdgeMarker = false;
		int grammarGrowthSteps = 1000;
		File lex = new File(args[0] + File.separator + "Lexicon" + args[1] + ".txt");
		File predLex = new File(args[0] + File.separator + "PredictionLexicon" + args[1] + ".txt");
		File outputDir = new File(args[2] + File.separator + "analysis" + args[1]);
		File grammarGrowthFile = new File(outputDir + File.separator + "analysis_grammar_growth.txt");
		outputDir.mkdir();
		Analysis ana = new Analysis(lex, predLex, extendedCategories, morphology, 
				secondaryEdgeMarker, grammarGrowthSteps, grammarGrowthFile);
		ana.analyse(outputDir, 50);
		
		if (args.length > 3) {
			String c = args[3];
			// coverage
			if (c.equals("-cov")) {
				File testFile = new File(args[4]);
				File testPredFile = null;
				if (testFile.isDirectory()) {
					File newTestFile = new File(args[4] + File.separator + "Lexicon" + testFile.getName() + ".txt");
					testFile = newTestFile;
					testPredFile = new File(args[4] + File.separator + "Prediction" + testFile.getName());
				}
				File covFile = new File(outputDir + File.separator + "analysis_coverage.txt");
				ana.coverage(testFile, testPredFile, covFile);
			}
		}
	}


	public Analysis(File lexFile, File predLexFile, boolean ext, boolean morph, boolean sec, int g, File gFile) {
		extendedCategories = ext;
		morphology = morph;
		secondaryEdgeMarker = sec;
		grammarGrowthSteps = g;
		growthFile = gFile;
		read(lexFile);		
		readPredLexUncombined(predLexFile);
	}
	
	public void analyse(File outputFolder, int max) {
		maxAnchors = max;
		
		try {
			
			argTreesFile = new File(outputFolder + File.separator + "analysis_initial.txt");
			modTreesFile = new File(outputFolder + File.separator + "analysis_modifier.txt");
			adjTreesFile = new File(outputFolder + File.separator + "analysis_auxiliary.txt");
			predTreesFile = new File(outputFolder + File.separator + "analysis_prediction.txt");
			ambiguityLexFile = new File(outputFolder + File.separator + "analysis_ambiguity_lexical.txt");
			ambiguityLemmaFile = new File(outputFolder + File.separator + "analysis_ambiguity_lemma.txt");
			templateFreqFile = new File(outputFolder + File.separator + "analysis_template_freq.txt");
			growthCoreFile = new File(outputFolder + File.separator + "analysis_core_grammar_growth.txt");
			matFile = new File(outputFolder + File.separator + "analysis_MATs.txt");
			sumFile = new File(outputFolder + File.separator + "analysis_summary.txt");
			lexiconFile = new File(outputFolder + File.separator + "analysis_lexicon.txt");
			
			ElementaryTreeComparator comparator = new ElementaryTreeComparator();
			WordFreqPairComparator wfpComparator = new WordFreqPairComparator();
			
			// sorting and printing argument trees
			for (String tree : argTrees.keySet()) {
				ElementaryTree elem = new ElementaryTree(tree, argTrees.get(tree));
				argTreesTokFreqSorted.add(elem);
			}
			Collections.sort(argTreesTokFreqSorted, comparator);
			
			fileWriter = new BufferedWriter(new FileWriter(argTreesFile));
			fileWriter.append("%%% Initial tree templates: " + argTrees.keySet().size() + "\n");
			printTrees(argTrees, argTreesTokFreqSorted);
			fileWriter.close();
			
			// sorting and printing modifier trees
			for (String tree : modTrees.keySet()) {
				ElementaryTree elem = new ElementaryTree(tree, modTrees.get(tree));
				modTreesTokFreqSorted.add(elem);
			}
			Collections.sort(modTreesTokFreqSorted, comparator);
			
			fileWriter = new BufferedWriter(new FileWriter(modTreesFile));
			fileWriter.append("%%% Modifier tree templates: " + modTrees.keySet().size() + "\n");
			printTrees(modTrees, modTreesTokFreqSorted);
			fileWriter.close();
			
			// sorting and printing auxiliary trees
			for (String tree : adjTrees.keySet()) {
				ElementaryTree elem = new ElementaryTree(tree, adjTrees.get(tree));
				adjTreesTokFreqSorted.add(elem);
			}
			Collections.sort(adjTreesTokFreqSorted, comparator);
			
			fileWriter = new BufferedWriter(new FileWriter(adjTreesFile));
			fileWriter.append("%%% Auxiliary tree templates: " + adjTrees.keySet().size() + "\n");
			printTrees(adjTrees, adjTreesTokFreqSorted);
			fileWriter.close();
			
			// sorting and printing prediction trees
			for (String tree : predTrees.keySet()) {
				ElementaryTree elem = new ElementaryTree(tree, predTrees.get(tree));
				predTreesTokFreqSorted.add(elem);
			}
			Collections.sort(predTreesTokFreqSorted, comparator);
			
			fileWriter = new BufferedWriter(new FileWriter(predTreesFile));
			fileWriter.append("%%% Prediction tree \"templates\": " + predTrees.keySet().size() + "\n");
			printPredTrees(predTrees, predTreesTokFreqSorted);
			fileWriter.close();
			
			// template frequencies: how often a template occurs in the data
			templatesTokFreqSorted.addAll(argTreesTokFreqSorted);
			templatesTokFreqSorted.addAll(modTreesTokFreqSorted);
			templatesTokFreqSorted.addAll(adjTreesTokFreqSorted);
			Collections.sort(templatesTokFreqSorted, comparator);
			fileWriter = new BufferedWriter(new FileWriter(templateFreqFile));
			fileWriter.append("rank\tfreq\ttemplate\n");
			int templFreqSum = 0; // = tokens
			int templFreqSumGT1000 = 0;
			int templFreqSum1 = 0;
			int countGT1000 = 0;
			int count1 = 0;
			for (int i = 0; i < templatesTokFreqSorted.size(); i++) {
				ElementaryTree e = templatesTokFreqSorted.get(i);
				int freq = e.getAnchorTokens().size();
				fileWriter.append(i+1 + "\t" + freq + "\t" + e.getElementaryTree() + "\n");
				templFreqSum += freq;
				if (freq >= 1000) {
					templFreqSumGT1000 += freq;
					countGT1000++;
				}
				if (freq == 1) {
					templFreqSum1 += freq;
					count1++;
				}
			}
			fileWriter.close();
			
			// etree types and etrees that occur only once
			int uniqEtrees = 0;
			int numEtrees = 0;
			for (String tree : argTrees.keySet()) {
				ArrayList<String> seenAnchors = stripLemmaOff(argTrees.get(tree));
				HashSet<String> seenAnchorsUniq = new HashSet(seenAnchors);
				numEtrees += seenAnchorsUniq.size();
				for (String a : seenAnchorsUniq) {
					if (seenAnchors.indexOf(a) == seenAnchors.lastIndexOf(a)) {
						// has been seen only once
						uniqEtrees++;
					}
				}
			}
			for (String tree : modTrees.keySet()) {
				ArrayList<String> seenAnchors = stripLemmaOff(modTrees.get(tree));
				HashSet<String> seenAnchorsUniq = new HashSet(seenAnchors);
				numEtrees += seenAnchorsUniq.size();
				for (String a : seenAnchorsUniq) {
					if (seenAnchors.indexOf(a) == seenAnchors.lastIndexOf(a)) {
						// has been seen only once
						uniqEtrees++;
					}
				}
			}
			for (String tree : adjTrees.keySet()) {
				ArrayList<String> seenAnchors = stripLemmaOff(adjTrees.get(tree));
				HashSet<String> seenAnchorsUniq = new HashSet(seenAnchors);
				numEtrees += seenAnchorsUniq.size();
				for (String a : seenAnchorsUniq) {
					if (seenAnchors.indexOf(a) == seenAnchors.lastIndexOf(a)) {
						// has been seen only once
						uniqEtrees++;
					}
				}
			}
			
			// ambiguity per word and per lemma
			int wordAmbigSum = 0;
			int lemmaAmbigSum = 0;
			
			for (String w : word2Tree.keySet()) {
				int ambig = word2Tree.get(w).size();
				WordFreqPair wfp = new WordFreqPair(w, ambig);
				word2TreeFreqSorted.add(wfp);
				wordAmbigSum += ambig;
			}
			Collections.sort(word2TreeFreqSorted, wfpComparator);
			double wordAmbig = (double)wordAmbigSum/word2TreeFreqSorted.size();
			fileWriter = new BufferedWriter(new FileWriter(ambiguityLexFile));
			fileWriter.append("rank\tfreq\tword\n");
			for (int i = 0; i < word2TreeFreqSorted.size(); i++) {
				WordFreqPair wfp = word2TreeFreqSorted.get(i);
				fileWriter.append(i+1 + "\t" + wfp.getFreq() + "\t" + wfp.getWord() + "\n");
				//System.out.println(word2Tree.get(wfp.getWord()));
			}
			fileWriter.close();
					
			for (String w : lemma2Tree.keySet()) {
				int ambig = lemma2Tree.get(w).size();
				WordFreqPair wfp = new WordFreqPair(w, ambig);
				lemma2TreeFreqSorted.add(wfp);
				lemmaAmbigSum += ambig;
			}
			Collections.sort(lemma2TreeFreqSorted, wfpComparator);
			double lemmaAmbig = (double)lemmaAmbigSum/lemma2TreeFreqSorted.size();
			fileWriter = new BufferedWriter(new FileWriter(ambiguityLemmaFile));
			fileWriter.append("rank\tfreq\tword\n");
			for (int i = 0; i < lemma2TreeFreqSorted.size(); i++) {
				WordFreqPair wfp = lemma2TreeFreqSorted.get(i);
				fileWriter.append(i+1 + "\t" + wfp.getFreq() + "\t" + wfp.getWord() + "\n");
			}
			fileWriter.close();
			
			int nPred = 1;
			int nCanon = 1;
			// growth of core grammar
			fileWriter = new BufferedWriter(new FileWriter(growthCoreFile));
			fileWriter.write("x\tcanonical\tprediction\tsum\n");
			int cLine = 0;
			for (WordFreqPair wfp : templateStorage) {
				cLine++;
				if (grammarGrowthSteps > 0 && cLine % grammarGrowthSteps == 0) {
					int sizeCanon = coreCanonLexicon.size();
					int sizePred = corePredLexicon.size();
					int sum = sizeCanon + sizePred;
					fileWriter.write(cLine + "\t" + sizeCanon + "\t" + sizePred + "\t" + sum + "\n");
				}
				String tree = wfp.getWord();
				if (wfp.getFreq() == 1) { // prediction
					int freq = predTrees.get(tree).size();
					if (freq > nPred && !corePredLexicon.contains(tree)) {
						corePredLexicon.add(tree); 
					}
				}
				else if (wfp.getFreq() == 0) { // canonical
					int freq = 0;
					if (argTrees.containsKey(tree)) {
						freq = argTrees.get(tree).size();
					}
					else if (modTrees.containsKey(tree)) {
						freq = modTrees.get(tree).size();
					}
					else if (adjTrees.containsKey(tree)) {
						freq = adjTrees.get(tree).size();
					}
					if (freq > nCanon && !coreCanonLexicon.contains(tree)) {
						coreCanonLexicon.add(tree);
					}
				}
			}
			if (grammarGrowthSteps > 0) {
				int sizeCanon = coreCanonLexicon.size();
				int sizePred = corePredLexicon.size();
				int sum = sizeCanon + sizePred;
				fileWriter.write(cLine + "\t" + sizeCanon + "\t" + sizePred + "\t" + sum + "\n");
			}
			fileWriter.close();
			
			int countPred = 0;
			for (Integer c : numTreesInPrecombinedPredTrees.values()) {
				countPred += c;
			}
			
			// MATs
			fileWriter = new BufferedWriter(new FileWriter(matFile));
			for (String t : MAT.keySet()) {
				fileWriter.append(t + "\n");
				fileWriter.append("\t\t" + MAT.get(t) + "\n");
			}
			fileWriter.close();
			
			// lexicon
			fileWriter = new BufferedWriter(new FileWriter(lexiconFile));
			for (String l : lexicon.keySet()) {
				for (String t : lexicon.get(l)) {
					fileWriter.append(l + "\t" + t + "\n");
				}
			}
			fileWriter.close();
			
			// summary
			fileWriter = new BufferedWriter(new FileWriter(sumFile));
			fileWriter.append("\n### Templates ###\n\n");
			fileWriter.append("Initial trees:\t" + argTrees.keySet().size() + "\n");
			fileWriter.append("Modifier trees:\t" + modTrees.keySet().size() + "\n");
			fileWriter.append("Auxiliary trees:\t" + adjTrees.keySet().size() + "\n");
			int sum = argTrees.keySet().size() + modTrees.keySet().size() + adjTrees.keySet().size();
			fileWriter.append("Sum:\t\t" + sum + "\n\n");
			fileWriter.append("Template frequency sum:\t" + templFreqSum + 
							"\t(number of template tokens = +1 for each extracted etree)\n");
			fileWriter.append(countGT1000 + " templates with freq >= 1000\t\taccount for " + 
							(double)templFreqSumGT1000/templFreqSum*100 + "% of all template tokens\n");
			fileWriter.append(count1 + " templates with freq == 1\t\taccount for " + 
					(double)templFreqSum1/templFreqSum*100 + "% of all template tokens\n\n");
			fileWriter.append("Combined Prediction trees:\t" + predTrees.keySet().size() + "\n");
			fileWriter.append("Uncombined prediction trees:\t" + predTreesUncombined.keySet().size() + "\n\n");
			
			fileWriter.append("Number of prediction trees in pre-combined trees: (sum " + countPred + ")\n");
			for (Integer i : numTreesInPrecombinedPredTrees.keySet()) {
				fileWriter.append("\t\t"+i+"\t" + 100*(double)numTreesInPrecombinedPredTrees.get(i)/countPred + "\n\n");
			}
			
			fileWriter.append("### Lexicalized elementary trees ###\n\n");
			fileWriter.append("Lexicalized elementary trees:\t" + numEtrees + "\n");
			fileWriter.append("Unique lexicalized elementary trees:\t" + uniqEtrees + "\t(seen only once)\n");
			fileWriter.append("\n### Lexical Ambiguity ###\n\n");
			fileWriter.append("Average ambiguity per word form:\t" + wordAmbig + " trees\n");
			fileWriter.append("Average ambiguity per lemma:\t" + lemmaAmbig + " trees\n\n");
			
			fileWriter.append("Number of MATs:\t" + MAT.size());
			fileWriter.close();
		}
		catch(IOException e) {
			e.printStackTrace();
		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	private void coverage(File testFile, File testPredFile, File covSummaryFile) {
		FileReader fr ;
		BufferedReader input = null;
		
		HashMap<String, ArrayList<String>> testPredTrees= new HashMap<String, ArrayList<String>>();
		HashMap<String, ArrayList<String>> testArgTrees= new HashMap<String, ArrayList<String>>();
		HashMap<String, ArrayList<String>> testModTrees= new HashMap<String, ArrayList<String>>();
		HashMap<String, ArrayList<String>> testAdjTrees= new HashMap<String, ArrayList<String>>();
		
		HashMap<String, ArrayList<String>> testPredTreesUncombined= new HashMap<String, ArrayList<String>>();
		
		// read the test file
		try {			
			fr = new FileReader(testFile);
			input = new BufferedReader(fr);
			
			String line = null;	
			while (( line = input.readLine()) != null){
				if (line.startsWith("%%%")) continue;
				String[] col = pTab.split(line);
				if (col.length < 3) {
					System.err.println("### Test file: not 3 or more columns: " + line);
					continue;
				}
				
				// prediction trees
				if (col[0].equals("prediction:")) {
					String tree = col[2];
					// delete extended categories if the flag is set accordingly
					if (!extendedCategories) {
						tree = deleteExtendedCategories(tree);
					}
					// delete morphology if the flag is set accordingly
					if (!morphology) {
						tree = deleteMorphology(tree);
					}
					
					// delete secondaryEdgeMarkers if the flat is set accordingly
					if (!secondaryEdgeMarker) {
						tree = deleteSecondaryEdgeMarkers(tree);
					}
					
					// replace trace identifiers
					tree = replaceTraceIdentifiers(tree);
					
					// delete filler information
					tree = deleteFillerInformation(tree);
					
					// make sure the origin identifiers are always in ascending order
					tree = putOriginIdentifiersInAscOrder(tree);					
					
					ArrayList<String> types = testPredTrees.get(tree);
					if (types == null) {
						types = new ArrayList<String>();
					}
					types.add(col[1]);
					testPredTrees.put(tree, types);
					continue;
				}
				
				// empty elements cannot be anchors in TAG
//				if (col[0].equals("- -")) {
//					if (isAnchoredInEmptyElement(col[2])) {
//						continue;
//					}
//				}
				
				// check whether there is exactly one anchor
				Matcher m = pAnchor.matcher(col[2]);
				int numAnch = 0;
				while (m.find() == true) {
					numAnch++;
				}
				if (numAnch != 1) {
					System.err.println("### Test file: not exactly one anchor: " + line);
					continue;
				}
				// replace anchor
				String tree = m.replaceFirst(" XXX<>");
				// replace trace identifiers
				tree = replaceTraceIdentifiers(tree);
				
				// delete extended categories if the flag is set accordingly
				if (!extendedCategories) {
					tree = deleteExtendedCategories(tree);
				}
				
				// delete morphology if the flag is set accordingly
				if (!morphology) {
					tree = deleteMorphology(tree);
				}
				
				// delete secondaryEdgeMarkers if the flat is set accordingly
				if (!secondaryEdgeMarker) {
					tree = deleteSecondaryEdgeMarkers(tree);
				}
				
				// delete filler information
				tree = deleteFillerInformation(tree);
				
				if (col[1].equals("ARG")) {
					ArrayList<String> anchors = testArgTrees.get(tree);
					if (anchors == null) {
						anchors = new ArrayList<String>();
					}
					anchors.add(col[0]);
					testArgTrees.put(tree, anchors);
				}
				else if (col[1].equals("MOD")) {
					ArrayList<String> anchors = testModTrees.get(tree);
					if (anchors == null) {
						anchors = new ArrayList<String>();
					}
					anchors.add(col[0]);
					testModTrees.put(tree, anchors);
				}
				else if (col[1].equals("ADJ")) {
					ArrayList<String> anchors = testAdjTrees.get(tree);
					if (anchors == null) {
						anchors = new ArrayList<String>();
					}
					anchors.add(col[0]);
					testAdjTrees.put(tree, anchors);
				}
				else {
					System.err.println("### Test file: what kind of tree?!: " + line);
				}
			}	
			
			// read lex pred file (uncombined prediction trees)
			fr = new FileReader(testPredFile);
			input = new BufferedReader(fr);
			
			line = null;	
			while (( line = input.readLine()) != null){
				String[] col = pTab.split(line);
				if (col.length != 4 && col.length != 3) {
					System.err.println("### Test file: not 3/4 columns: " + line);
					continue;
				}
				
				// prediction trees
				if (col[0].equals("prediction:")) {
					String tree = col[2];
					// delete extended categories if the flag is set accordingly
					if (!extendedCategories) {
						tree = deleteExtendedCategories(tree);
					}
					// delete morphology if the flag is set accordingly
					if (!morphology) {
						tree = deleteMorphology(tree);
					}
					
					// delete secondaryEdgeMarkers if the flat is set accordingly
					if (!secondaryEdgeMarker) {
						tree = deleteSecondaryEdgeMarkers(tree);
					}
					
					// replace trace identifiers
					tree = replaceTraceIdentifiers(tree);
				
					ArrayList<String> types = testPredTreesUncombined.get(tree);
					if (types == null) {
						types = new ArrayList<String>();
					}
					types.add(col[1]);
					testPredTreesUncombined.put(tree, types);
				}
			}
		}
		catch (FileNotFoundException e) {
			System.err.println(e.getMessage());
			//e.printStackTrace();
		}
		catch (IOException ex){
			ex.printStackTrace();
		}
		finally {
			try {
				if (input!= null) {
					input.close();
				}
			}
			catch (IOException ex) {
				ex.printStackTrace();
			}
		}
		
		// check coverage
		int numPred = 0;
		int numPredSeen = 0;
		int numPredTok = 0;
		int numPredTokSeen = 0;
		int numTempl = 0;
		int numTemplSeen = 0;
		int numTemplTok = 0;
		int numTemplTokSeen = 0;
		int numEtreeTok = 0;
		int numEtreeTokSeen = 0;
		int numEtreeTokOutDict = 0;
		int numLemmaEtreeTok = 0;
		int numLemmaEtreeTokSeen = 0;
		int numLemmaEtreeTokOutDict = 0;
		ArrayList<String> uncovered = new ArrayList<String>();
		
		for (String tree : testPredTrees.keySet()) {
			// types
			numPred++;
			if (predTrees.containsKey(tree)) {
				numPredSeen++;
			}
			else {
				uncovered.add(testPredTrees.get(tree).size() + "\t" + tree);
			}
			//else {
			//	System.out.println(tree);
			//}
			// tokens
			for (String t : testPredTrees.get(tree)) {
				numPredTok++;
				if (predTrees.containsKey(tree)) {
					numPredTokSeen++;
				}
			}
		}
		for (String tree : testArgTrees.keySet()) {
			// template types
			numTempl++;
			if (argTrees.containsKey(tree)) {
				numTemplSeen++;
			}
			else {
				uncovered.add(testArgTrees.get(tree).size() + "\t" + tree + "\t" + testArgTrees.get(tree));
			}
			// template tokens
			for (String t : testArgTrees.get(tree)){
				numTemplTok++;
				if (argTrees.containsKey(tree)) {
					numTemplTokSeen++;
				}
			}
			// (lexicalized) etree tokens
			ArrayList<String> anchors = stripLemmaOff(testArgTrees.get(tree));
			ArrayList<String> seenAnchors = argTrees.get(tree);
			if (seenAnchors == null) {
				seenAnchors = new ArrayList<String>();
			}
			seenAnchors = stripLemmaOff(seenAnchors);
			for (String anch : anchors) {
				numEtreeTok++;
				if (seenAnchors.contains(anch)) {
					numEtreeTokSeen++;
				}
				else {
					// check whether the word has been seen in the training data at all
					if (word2Tree.get(anch) == null) {
						numEtreeTokOutDict++;
					}
				}
			}
			// lemma-lexicalized etree tokens
			ArrayList<String> anchorsLemma = stripWordFormOff(testArgTrees.get(tree));
			ArrayList<String> seenAnchorsLemma = argTrees.get(tree);
			if (seenAnchorsLemma == null) {
				seenAnchorsLemma = new ArrayList<String>();
			}
			seenAnchorsLemma = stripWordFormOff(seenAnchorsLemma);
			for (String anchLemma : anchorsLemma) {
				numLemmaEtreeTok++;
				if (seenAnchorsLemma.contains(anchLemma)) {
					numLemmaEtreeTokSeen++;
				}
				else {
					// check whether the lemma has been seen n the training data at all
					if (lemma2Tree.get(anchLemma) == null) {
						numLemmaEtreeTokOutDict++;
					}
				}
			}
		}
		for (String tree : testModTrees.keySet()) {
			// template types
			numTempl++;
			if (modTrees.containsKey(tree)) {
				numTemplSeen++;
			}
			else {
				uncovered.add(testModTrees.get(tree).size() + "\t" + tree + "\t" + testModTrees.get(tree));
			}
			// template tokens
			for (String t : testModTrees.get(tree)){
				numTemplTok++;
				if (modTrees.containsKey(tree)) {
					numTemplTokSeen++;
				}
			}
			// (lexicalized) etree tokens
			ArrayList<String> anchors = stripLemmaOff(testModTrees.get(tree));
			ArrayList<String> seenAnchors = modTrees.get(tree);
			if (seenAnchors == null) {
				seenAnchors = new ArrayList<String>();
			}
			seenAnchors = stripLemmaOff(seenAnchors);
			for (String anch : anchors) {
				numEtreeTok++;
				if (seenAnchors.contains(anch)) {
					numEtreeTokSeen++;
				}
				else {
					// check whether the word has been seen in the training data at all
					if (word2Tree.get(anch) == null) {
						numEtreeTokOutDict++;
					}
				}
			}
			// lemma-lexicalized etree tokens
			ArrayList<String> anchorsLemma = stripWordFormOff(testModTrees.get(tree));
			ArrayList<String> seenAnchorsLemma = modTrees.get(tree);
			if (seenAnchorsLemma == null) {
				seenAnchorsLemma = new ArrayList<String>();
			}
			seenAnchorsLemma = stripWordFormOff(seenAnchorsLemma);
			for (String anchLemma : anchorsLemma) {
				numLemmaEtreeTok++;
				if (seenAnchorsLemma.contains(anchLemma)) {
					numLemmaEtreeTokSeen++;
				}
				else {
					// check whether the lemma has been seen n the training data at all
					if (lemma2Tree.get(anchLemma) == null) {
						numLemmaEtreeTokOutDict++;
					}
				}
			}
		}
		for (String tree : testAdjTrees.keySet()) {
			// template types
			numTempl++;
			if (adjTrees.containsKey(tree)) {
				numTemplSeen++;
			}
			else {
				uncovered.add(testAdjTrees.get(tree).size() + "\t" + tree + "\t" + testAdjTrees.get(tree));
			}
			// template tokens
			for (String t : testAdjTrees.get(tree)){
				numTemplTok++;
				if (adjTrees.containsKey(tree)) {
					numTemplTokSeen++;
				}
			}
			// (lexicalized) etree tokens
			ArrayList<String> anchors = stripLemmaOff(testAdjTrees.get(tree));
			ArrayList<String> seenAnchors = adjTrees.get(tree);
			if (seenAnchors == null) {
				seenAnchors = new ArrayList<String>();
			}
			seenAnchors = stripLemmaOff(seenAnchors);
			for (String anch : anchors) {
				numEtreeTok++;
				if (seenAnchors.contains(anch)) {
					numEtreeTokSeen++;
				}
				else {
					// check whether the word has been seen in the training data at all
					if (word2Tree.get(anch) == null) {
						numEtreeTokOutDict++;
					}
				}
			}
			// lemma-lexicalized etree tokens
			ArrayList<String> anchorsLemma = stripWordFormOff(testAdjTrees.get(tree));
			ArrayList<String> seenAnchorsLemma = adjTrees.get(tree);
			if (seenAnchorsLemma == null) {
				seenAnchorsLemma = new ArrayList<String>();
			}
			seenAnchorsLemma = stripWordFormOff(seenAnchorsLemma);
			for (String anchLemma : anchorsLemma) {
				numLemmaEtreeTok++;
				if (seenAnchorsLemma.contains(anchLemma)) {
					numLemmaEtreeTokSeen++;
				}
				else {
					// check whether the lemma has been seen n the training data at all
					if (lemma2Tree.get(anchLemma) == null) {
						numLemmaEtreeTokOutDict++;
					}
				}
			}
		}
		// uncombined prediction trees
		for (String tree : testPredTreesUncombined.keySet()) {
			if (!predTreesUncombined.containsKey(tree)) {
				uncovered.add("UNCOMBINED\t" + tree);
			}
		}
		
		BufferedWriter coverageWriter = null;
		try {
			coverageWriter = new BufferedWriter(new FileWriter(covSummaryFile));
			coverageWriter.write("COVERAGE\n\n");
			coverageWriter.write("Template types covered: " + (double)numTemplSeen/numTempl + "\n");	
			coverageWriter.write("Template tokens covered: " + (double)numTemplTokSeen/numTemplTok + "\n\n");
			coverageWriter.write("Etree tokens covered (lexicalization based on word form): " + 
							(double)numEtreeTokSeen/numEtreeTok + "\n");
			coverageWriter.write("Etree tokens not covered & out of dict: " + 
					(double)numEtreeTokOutDict/numEtreeTok + "\n\n");
			coverageWriter.write("Etree tokens covered (lexicalization based on lemma): " + 
							(double)numLemmaEtreeTokSeen/numLemmaEtreeTok + "\n");
			coverageWriter.write("Etree tokens not covered (lexicalization based on lemma) & out of lemma-dict: " + 
					(double)numEtreeTokOutDict/numLemmaEtreeTok + "\n\n");
			coverageWriter.write("Prediction tree types covered: " + (double)numPredSeen/numPred + "\n");
			coverageWriter.write("Prediction tree tokens covered: " + (double)numPredTokSeen/numPredTok + "\n\n");
			coverageWriter.write("Template types covered (including prediction trees): " + 
							((double)numPredSeen + (double)numTemplSeen)/(numPred + numTempl) + "\n");
			coverageWriter.write("Template tokens covered (including prediction trees): " + 
							((double)numPredTokSeen + (double)numTemplTokSeen)/(numPredTok + numTemplTok) + "\n");
		
			coverageWriter.write("\n\nUncovered trees\n\n");
			for (String t : uncovered) {
				coverageWriter.write(t + "\n");
			}
		}
		catch (FileNotFoundException e) {
			System.err.println(e.getMessage());
			//e.printStackTrace();
		}
		catch (IOException ex){
			ex.printStackTrace();
		}
		finally {
			try {
				if (coverageWriter != null) {
					coverageWriter.close();
				}
			}
			catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}



	private void readPredLexUncombined(File predLexFile) {
		FileReader fr;
		BufferedReader input = null;
		try {			
			fr = new FileReader(predLexFile);
			input = new BufferedReader(fr);
			
			String line = null;	
			while (( line = input.readLine()) != null) {
				String[] col = pTab.split(line);
				if (col.length != 4 && col.length != 3) {
					System.err.println("### not 3/4 columns: " + line);
					continue;
				}
				
				// prediction trees
				if (col[0].equals("prediction:")) {
					String tree = col[2];
					// delete extended categories if the flag is set accordingly
					if (!extendedCategories) {
						Matcher m2 = pExtCat2.matcher(tree);
						tree = m2.replaceAll("");
					}
					// delete morphology if the flag is set accordingly
					if (!morphology) {
						Matcher m2 = pMorph.matcher(tree);
						tree = m2.replaceAll("");
					}
					
					// delete secondaryEdgeMarkers if the flat is set accordingly
					if (!secondaryEdgeMarker) {
						Matcher m2 = pSecondaryEdgeMarker.matcher(tree);
						tree = m2.replaceAll("");
					}
					
					// replace trace identifiers
					Matcher m3 = pTrace.matcher(tree);
					tree = m3.replaceAll("*T1*");
					
					// delete filler information
					tree = deleteFillerInformation(tree);
				
					ArrayList<String> types = predTreesUncombined.get(tree);
					if (types == null) {
						types = new ArrayList<String>();
					}
					types.add(col[1]);
					predTreesUncombined.put(tree, types);
					
				}
			}
		}
		catch (FileNotFoundException e) {
			System.err.println(e.getMessage());
		}
		catch (IOException ex){
			ex.printStackTrace();
		}
		finally {
			try {
				if (input!= null) {
					input.close();
				}
			}
			catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}


	private void read(File lexFile) {
		FileReader fr ;
		BufferedReader input = null;
		BufferedWriter growthWriter = null;
		
		try {			
			fr = new FileReader(lexFile);
			input = new BufferedReader(fr);
			growthWriter = new BufferedWriter(new FileWriter(growthFile));
			growthWriter.write("x\tcanonical\tprediction\tsum\n");
			
			String line = null;	
			int cLine = 0;
			while (( line = input.readLine()) != null){
				if (line.startsWith("%%%")) continue;
				cLine++;
				String[] col = pTab.split(line);
				if (col.length < 3) {
					System.err.println("### not 3 or more columns: " + line);
					continue;
				}
				
				// grammar growth
				if (grammarGrowthSteps > 0 && cLine % grammarGrowthSteps == 0) {
					int sizeCanon = argTrees.size() + modTrees.size() + adjTrees.size();
					int sizePred = predTrees.size();
					int sum = sizeCanon + sizePred;
					growthWriter.write(cLine + "\t" + sizeCanon + "\t" + sizePred + "\t" + sum + "\n");
				}
				
				// prediction trees
				if (col[0].equals("prediction:")) {
					String tree = col[2];
					// delete extended categories if the flag is set accordingly
					if (!extendedCategories) {
						Matcher m2 = pExtCat2.matcher(tree);
						tree = m2.replaceAll("");
					}
					// delete morphology if the flag is set accordingly
					if (!morphology) {
						Matcher m2 = pMorph.matcher(tree);
						tree = m2.replaceAll("");
					}
					
					// delete secondaryEdgeMarkers if the flat is set accordingly
					if (!secondaryEdgeMarker) {
						Matcher m2 = pSecondaryEdgeMarker.matcher(tree);
						tree = m2.replaceAll("");
					}
					
					// replace trace identifiers
					Matcher m3 = pTrace.matcher(tree);
					tree = m3.replaceAll("*T1*");
					
					// delete filler information
					tree = deleteFillerInformation(tree);
					
					// make sure the origin identifiers are always in ascending order
					int i = 0;
					HashMap<Integer, Integer> indexMapping = new HashMap<Integer, Integer>();
					Matcher m4 = pIndex.matcher(tree);
					StringBuffer sb = new StringBuffer();
					while (m4.find()) {
						int ind = Integer.parseInt(m4.group(2));
						if (!indexMapping.containsKey(ind)) {
							i++;
							indexMapping.put(ind, i);
						}
						m4.appendReplacement(sb, m4.group(1)+indexMapping.get(ind));
					}
					m4.appendTail(sb);
					tree = sb.toString();
					putNumPredTrees(i);
					
					
					ArrayList<String> types = predTrees.get(tree);
					if (types == null) {
						types = new ArrayList<String>();
					}
					types.add(col[1]);
					predTrees.put(tree, types);
					// for the core grammar computation
					templateStorage.add(new WordFreqPair(tree, 1));
					continue;
				}
				// MATs
				Matcher mMAT = pCoAnchor.matcher(col[2]);
				if (mMAT.find()) {
					recordMAT(col[2]);
				}
				
				// empty elements cannot be anchors in TAG
//				if (col[0].equals("- -")) {
//					if (isAnchoredInEmptyElement(col[2])) {
//						continue;
//					}
//				}
				
				// check whether there is exactly one anchor
				Matcher m = pAnchor.matcher(col[2]);
				int numAnch = 0;
				while (m.find() == true) {
					numAnch++;
				}
				if (numAnch != 1) {
					System.err.println("### not exactly one anchor: " + line);
					continue;
				}
				// replace anchor
				String tree = m.replaceFirst(" XXX<>");
				
				// replace trace identifiers
				Matcher m3 = pTrace.matcher(tree);
				tree = m3.replaceAll("*T1*");
				
				// delete extended categories if the flag is set accordingly
				if (!extendedCategories) {
					Matcher m2 = pExtCat2.matcher(tree);
					tree = m2.replaceAll("");
				}
				
				// delete morphology if the flag is set accordingly
				if (!morphology) {
					Matcher m2 = pMorph.matcher(tree);
					tree = m2.replaceAll("");
				}
				
				// delete secondaryEdgeMarkers if the flat is set accordingly
				if (!secondaryEdgeMarker) {
					Matcher m2 = pSecondaryEdgeMarker.matcher(tree);
					tree = m2.replaceAll("");
				}
				
				// delete filler information
				tree = deleteFillerInformation(tree);
				
				if (col[1].equals("ARG")) {
					ArrayList<String> anchors = argTrees.get(tree);
					if (anchors == null) {
						anchors = new ArrayList<String>();
					}
					anchors.add(col[0]);
					argTrees.put(tree, anchors);
				}
				else if (col[1].equals("MOD")) {
					ArrayList<String> anchors = modTrees.get(tree);
					if (anchors == null) {
						anchors = new ArrayList<String>();
					}
					anchors.add(col[0]);
					modTrees.put(tree, anchors);
				}
				else if (col[1].equals("ADJ")) {
					ArrayList<String> anchors = adjTrees.get(tree);
					if (anchors == null) {
						anchors = new ArrayList<String>();
					}
					anchors.add(col[0]);
					adjTrees.put(tree, anchors);
				}
				else {
					System.err.println("### what kind of tree?!: " + line);
				}
				
				// lexicon
				add2Lexicon(col[0], tree);
				
				// ambiguity
				String[] w = pWhite.split(col[0]);
				ArrayList<String> templ1 = word2Tree.get(w[0]);
				if (templ1 == null) {
					templ1 = new ArrayList<String>();
				}
				if (!templ1.contains(tree)) {
					templ1.add(tree);
				}
				word2Tree.put(w[0], templ1);
				ArrayList<String> templ2 = lemma2Tree.get(w[1]);
				if (templ2 == null) {
					templ2 = new ArrayList<String>();
				}
				if (!templ2.contains(tree)) {
					templ2.add(tree);
				}
				lemma2Tree.put(w[1], templ2);
				
				// for the core grammar computation
				templateStorage.add(new WordFreqPair(tree, 0));
			}	
			// grammar growth (for last line)
			if (grammarGrowthSteps > 0) {
				int sizeCanon = argTrees.size() + modTrees.size() + adjTrees.size();
				int sizePred = predTrees.size();
				int sum = sizeCanon + sizePred;
				growthWriter.write(cLine + "\t" + sizeCanon + "\t" + sizePred + "\t" + sum + "\n");
			}
		}
		catch (FileNotFoundException e) {
			System.err.println(e.getMessage());
			//e.printStackTrace();
		}
		catch (IOException ex){
			ex.printStackTrace();
		}
		finally {
			try {
				if (input!= null) {
					input.close();
				}
				growthWriter.close();
			}
			catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}




	private void recordMAT(String tree) {
		// replace trace identifiers
		tree = replaceTraceIdentifiers(tree);
		
		// delete extended categories if the flag is set accordingly
		if (!extendedCategories) {
			tree = deleteExtendedCategories(tree);
		}
		
		// delete morphology if the flag is set accordingly
		if (!morphology) {
			tree = deleteMorphology(tree);
		}
		
		// delete secondaryEdgeMarkers if the flat is set accordingly
		if (!secondaryEdgeMarker) {
			tree = deleteSecondaryEdgeMarkers(tree);
		}
		
		if (MAT.containsKey(tree)) {
			MAT.put(tree, MAT.get(tree)+1);
		}
		else {
			MAT.put(tree, 1);
		}
		
	}


	private void putNumPredTrees(int i) {
		Integer count = numTreesInPrecombinedPredTrees.get(i);
		if (count == null) {
			numTreesInPrecombinedPredTrees.put(i, 1);
		}
		else {
			numTreesInPrecombinedPredTrees.put(i, count+1);
		}
		
	}


	private void printTrees(HashMap<String, ArrayList<String>> trees,
			List<ElementaryTree> freq) throws IOException {
		
		HashSet<String> types;
		for (ElementaryTree tree : freq) {	
			fileWriter.append(tree.getElementaryTree() + "\n%%% seen " + tree.getAnchorTokens().size() 
					+ " times, " + tree.getAnchorTypes().size() + " different types\n");
			int i = 0;
			for (String anchor : tree.getAnchorTypes()) {
				fileWriter.append("\t" + anchor + "\n");
				i++;
				if (i >= maxAnchors) {
					break;
				}
			}
		}
	}
	
	private void printPredTrees(HashMap<String, ArrayList<String>> predTrees,
			List<ElementaryTree> predTreesTokFreqSorted) throws IOException {
		for (ElementaryTree tree : predTreesTokFreqSorted) {
			fileWriter.append(tree.getElementaryTree() + "\n");
			for (String type : tree.getAnchorTypes()) {
				fileWriter.append("\t" + type);
			}
			fileWriter.append("  %%% seen " + tree.getAnchorTokens().size() + " times\n");
		}
		
	}

	
	private String deleteExtendedCategories(String tree) {
		Matcher m2 = pExtCat2.matcher(tree);
		tree = m2.replaceAll("");
		return tree;
	}
	

	private String deleteMorphology(String tree) {
		Matcher m2 = pMorph.matcher(tree);
		tree = m2.replaceAll("");
		return tree;
	}
	

	private String deleteSecondaryEdgeMarkers(String tree) {
		Matcher m2 = pSecondaryEdgeMarker.matcher(tree);
		tree = m2.replaceAll("");	
		return tree;
	}
	
	private boolean isAnchoredInEmptyElement(String tree) {
		Matcher m3 = pTraceTree.matcher(tree);
		return m3.find();
	}
	
	private String replaceTraceIdentifiers(String tree) {
		Matcher m3 = pTrace.matcher(tree);
		tree = m3.replaceAll("*T1*");
		return tree;
	}
	
	private String deleteFillerInformation(String tree) {
		Matcher m3 = pFiller.matcher(tree);
		tree = m3.replaceAll("");
		return tree;
	}
	
	private String putOriginIdentifiersInAscOrder(String tree) {
		// make sure the origin identifiers are always in ascending order
		int i = 0;
		HashMap<Integer, Integer> indexMapping = new HashMap<Integer, Integer>();
		Matcher m4 = pIndex.matcher(tree);
		StringBuffer sb = new StringBuffer();
		while (m4.find()) {
			int ind = Integer.parseInt(m4.group(2));
			if (!indexMapping.containsKey(ind)) {
				i++;
				indexMapping.put(ind, i);
			}
			m4.appendReplacement(sb, m4.group(1)+indexMapping.get(ind));
		}
		m4.appendTail(sb);
		tree = sb.toString();		
		return tree;
	}
	
	private ArrayList<String> stripLemmaOff(ArrayList<String> arrayList) {
		ArrayList<String> anch = new ArrayList<String>();
		for (String pair : arrayList) {
			String[] col = pWhite.split(pair);
			if (col.length >= 2) {
				anch.add(col[0]);
			}
		}
		return anch;
	}
	
	private ArrayList<String> stripWordFormOff(ArrayList<String> arrayList) {
		ArrayList<String> anch = new ArrayList<String>();
		for (String pair : arrayList) {
			String[] col = pWhite.split(pair);
			if (col.length >= 2) {
				anch.add(col[1]);
			}
		}
		return anch;
	}
	
	private void add2Lexicon(String l, String t) {
		ArrayList<String> list = lexicon.get(l);
		if (list == null) {
			list = new ArrayList<String>();
		}
		if (!list.contains(t)) {
			list.add(t);
		}
		lexicon.put(l, list);
	}
}

class ElementaryTreeComparator implements Comparator<ElementaryTree> {
	public int compare(ElementaryTree l1, ElementaryTree l2) {
		int r1 = l1.getAnchorTokens().size();
		int r2 = l2.getAnchorTokens().size();
		if (r1 > r2) {
			return -1;
		}
		else if (r1 < r2) {
			return 1;
		}
		else {
			return 0;
		}
	}
}

class WordFreqPairComparator implements Comparator<WordFreqPair> {
	public int compare(WordFreqPair l1, WordFreqPair l2) {
		int r1 = l1.getFreq();
		int r2 = l2.getFreq();
		if (r1 > r2) {
			return -1;
		}
		else if (r1 < r2) {
			return 1;
		}
		else {
			return 0;
		}
	}
}

class WordFreqPair {
	private String word;
	private int freq;
	
	public WordFreqPair(String w, int f) {
		word = w;
		freq = f;
	}
	
	public String getWord() {
		return word;
	}
	
	public int getFreq() {
		return freq;
	}
}
