package pltag.german.regularadjunction;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import pltag.german.AbstractPredictionTreeExtractor;
import pltag.german.CompositeTree;
import pltag.german.ElementaryTree;
import pltag.german.LeafNode;
import pltag.german.LexEntry;
import pltag.german.NodeStatus;
import pltag.german.PredictionTree;
import pltag.german.PredictionTreeContainer;
import pltag.german.TreeNode;
import pltag.german.Utils;



public class PredictionTreeExtractorRegAdj extends
		AbstractPredictionTreeExtractor {
	
	private ArrayList<TreeNode> connectionPath = new ArrayList<TreeNode>();
	private TreeNode connectionPathRoot = null;
	
	private HashMap<ElementaryTree, HashSet<TreeNode>> nodesRequired;
	private ArrayList<PredictionTreeContainer> futurePredTrees;

	public PredictionTreeExtractorRegAdj(boolean mat) {
		super(mat);
		// TODO Auto-generated constructor stub
	}

	@Override
	protected void extract() {
		connectionPath = new ArrayList<TreeNode>();
		connectionPathRoot = null;
		nodesRequired = new HashMap<ElementaryTree,HashSet<TreeNode>>();
		futurePredTrees = new ArrayList<PredictionTreeContainer>();
		
		computeConnectionPaths(currTree.getRoot());
		buildPredTrees();
	}
	
	
	private void computeConnectionPaths(TreeNode node) {
		
		if (node.isLeaf()) {
		
			connectionPath.add(node);
			node.getBottom().markNodesAsPredicted(); // how about traces? and MAT?
			
			// connectionPath now contains all nodes on the path from the previous word to the current word
			if (node.getID() != 0) { // not for the first word
				computePredTree(node);
			}
			//clearing
			connectionPath.clear();
			connectionPathRoot = null;
			// starting a new connectionPath
			connectionPath.add(node);

			return;
		}
		else {
			connectionPath.add(node);
			for (TreeNode child : node.getChildren()) {
				// basically we just jump over traces.
				// this has the effect that no prediction tree is generated
				// due to some empty element (whether bound or not).
				// bound traces will be introduced "afterwards" with their elementary tree.
				// we don't care about unbound traces.
				if (child.isLeaf() && ((LeafNode)child).isTrace()) {
					continue;
				}
				computeConnectionPaths(child);
				connectionPath.add(node);
				connectionPathRoot = node;
			}
		}
		
	}

	private void computePredTree(TreeNode node) {
		int i = node.getID();
		
		ArrayList<ElementaryTree> eTreesRequired = new ArrayList<ElementaryTree>();
		ArrayList<LexEntry> correspondingLexEntries = new ArrayList<LexEntry>();
		
		HashSet<ElementaryTree> eTreesToRemove = new HashSet<ElementaryTree>();
		
		for (int c = 0; c < connectionPath.size(); c++) {
			TreeNode cand = connectionPath.get(c);
			if (cand.isLeaf() && ((LeafNode)cand).isTrace()) {
				// how to treat them?!
				continue;
			}
			
			// TOP
			if (!cand.isPredictedTop() && cand.getTop() != null && cand != connectionPathRoot) {
				
				LexEntry currLex = cand.getTop();
				ElementaryTree eTree = currLex.getElementaryTree();
				if (eTreesRequired.contains(eTree)
						|| nodesRequired.containsKey(eTree)) {				
					// keep track of the nodes that require this prediction tree
					addNodeToRequiredNodes(cand, eTree);
					
				}
				else if (cand.getTop().isAdjTree() && cand == cand.getTop().getFoot()) {
					
				}
				else {
					eTreesRequired.add(eTree);
					correspondingLexEntries.add(currLex);
					// keep track of the nodes that require this prediction tree
					addNodeToRequiredNodes(cand, eTree);
				}
			}
			
			// BOTTOM
			if (!cand.isPredictedBottom() && cand.getBottom() != null) {
				
				LexEntry currLex = cand.getBottom();
				ElementaryTree eTree = currLex.getElementaryTree();
				TreeNode candNextOnPath = connectionPath.get(c+1);
				
				//TreeNode topOfAdjChain = findTopOfAdjChain(cand);
				
				if (eTreesRequired.contains(eTree)
						|| nodesRequired.containsKey(eTree)) {
					
					// keep track of the nodes that require this prediction tree
					addNodeToRequiredNodes(cand, eTree);
				}

				
				else if (eTree.getType() == NodeStatus.ADJ) {
					// the bottom of a node has not yet been predicted.
					// this might be an auxiliary tree that we can factor out and thus
					// it should not be predicted.
					// however, we might have to predict it, if more than the spine
					// is on the connection path, e.g. a substitution node, OR
					// if the node we are just computing the prediction tree for
					// will sister-adjoin to a node of the aux tree (not to the foot node)
					// in order to predict the correct attachment, generating a 
					// prediction tree for the aux tree is necessary.
					
					TreeNode foot = null;
					TreeNode root = null;
					HashSet<TreeNode> others = new HashSet<TreeNode>();				
					// check which other nodes that belong to this aux tree
					// are on the connection path
					for (TreeNode otherNode : connectionPath) {
						if (otherNode.getTop() == currLex &&
									currLex.getFoot() == otherNode) {
							foot = otherNode;
						}
						else if (otherNode.getBottom() == currLex &&
									otherNode == currLex.getRoot()) {
							root = otherNode;
						}
						else if (otherNode.getTop() == currLex || 
									otherNode.getBottom() == currLex) {
							others.add(otherNode);							
						}
					}
					boolean predictedAdj = false;
					for (TreeNode other : others) {
						if (other.getTop() == currLex && other.getBottom() != currLex) {
							// A node other than root and foot is on the connection path.
							// We predict the aux tree
							eTreesRequired.add(eTree);
							correspondingLexEntries.add(currLex);
							
							// keep track of the predicted nodes
							// and of the nodes that require this prediction tree
							addNodeToRequiredNodes(root, eTree);
							addNodeToRequiredNodes(foot, eTree);
							for (TreeNode other2: others) {
								addNodeToRequiredNodes(other2, eTree);
							}
							
							predictedAdj = true;
							break;
						}
					}
					if (!predictedAdj) {
						if (root != null && foot != null) {
							// does not need to be predicted
							eTreesRequired.add(eTree);
							correspondingLexEntries.add(currLex);
							eTreesToRemove.add(eTree);
						}
						else if (root != null && foot == null) {
							// only the root is only prediction path. (its top has not been predicted.)
							// we should predict it here.
							eTreesRequired.add(eTree);
							correspondingLexEntries.add(currLex);
							//setPredicted(root, currLex);
							addNodeToRequiredNodes(root, eTree);
						}
						else {
							currErrorString.append(" weird case, skipping adj tree " + Utils.subtreeToString(node));
						}
					}
				}
				else {
					eTreesRequired.add(eTree);
					correspondingLexEntries.add(currLex);
					addNodeToRequiredNodes(cand, eTree);
				}
			}			
			
		}

		
		for (ElementaryTree toRemove : eTreesToRemove) {
			int k = eTreesRequired.indexOf(toRemove);
			eTreesRequired.remove(k);
			correspondingLexEntries.remove(k);
			nodesRequired.remove(toRemove);
		}
		
		if (eTreesRequired.size() == 0) {
			// no prediction tree is needed for connectivity
			return;
		}
		
		for (int k = 0; k < eTreesRequired.size(); k++) {
			ElementaryTree eTree = eTreesRequired.get(k);
			LexEntry l = correspondingLexEntries.get(k);
			
			PredictionTreeContainer container = new PredictionTreeContainer(eTree, l, i);
			futurePredTrees.add(container);
		}
		
	}
	
	private void buildPredTrees() {
		
		ArrayList<PredictionTree> preds = new ArrayList<PredictionTree>();
		
		for (int k = 0; k < futurePredTrees.size(); k++) {
			PredictionTreeContainer container = futurePredTrees.get(k);
			int i = container.getIndex();
			PredictionTree predTree = new PredictionTree(i, container.getElementaryTree());
			predTree.cutUnary(predTree.getRoot(), nodesRequired.get(container.getElementaryTree()));
			preds.add(predTree);
			
		
			if (futurePredTrees.size()-1 == k 
					|| futurePredTrees.get(k+1).getIndex() != i) {
				// combine the prediction trees at the current index
				if (preds.size() > 1) {
					CompositeTree partialTree = new CompositeTree(preds.get(0));
					StringBuffer errorString = new StringBuffer();
					for (int j = 1; j < preds.size(); j++) {
						partialTree.tryToIntegratePrediction(preds.get(j), errorString, true);
						//System.out.println(partialTree.toString());
					}
					if (errorString.length() == 0) {
						boolean success = partialTree.tryMorePrediction(errorString, true);
						if (!success) {
							errorString.append(" Error in combining prediction trees." + preds);
						}
					}
					if (errorString.length() != 0) {
						currErrorString.append(errorString);
					}
					else {
						// we have combined all prediction trees into one prediction tree at one index
						PredictionTree combinedTree = new PredictionTree(i, partialTree);
						preds.clear();
						preds.add(combinedTree);
					}
				}
				predLex.addAll(preds);
				preds.clear();
			}			
		}
	}
	
	private void addNodeToRequiredNodes(TreeNode node, ElementaryTree eTree) {
		
		HashSet<TreeNode> required = nodesRequired.get(eTree);
		if (required == null) {
			required = new HashSet<TreeNode>();
		}
		required.add(node);
		nodesRequired.put(eTree, required);
	}

}
