package pltag.german.sisteradjunction;

import java.util.ArrayList;

import pltag.german.AbstractGrammarExtractor;
import pltag.german.LeafNode;
import pltag.german.LexEntry;
import pltag.german.NodeStatus;
import pltag.german.TreeNode;
import pltag.german.Utils;

public class GrammarExtractorSisAdj extends AbstractGrammarExtractor {
	


	public GrammarExtractorSisAdj(boolean multiAnchoredTrees) {
		super(multiAnchoredTrees);
	}

	protected void extract(TreeNode node, LexEntry lex) {
		LeafNode leaf = extractSubtree(node, lex);
		lex.setLeafLink(leaf);
		currLexicon.add(lex);
	}
	
	private TreeNode extractPredAdj(TreeNode node, LexEntry lex) {
		lex.setAdjTree();
		LeafNode leaf = extractSubtree(node, lex, false);
		lex.setLeafLink(leaf);
		currLexicon.add(lex);
		return lex.getFoot();
	}
	
	private void addToLexEntry(TreeNode node, LexEntry lex) {
		extractSubtree(node, lex);
		// here we only add to the lex entry, but we do not put it
		// to the lexicon
	}
	
	private LeafNode extractSubtree(TreeNode node, LexEntry lex) {
		return extractSubtree(node, lex, true);
	}
	
	// the boolean is needed such that extractPredAdj does not go into an endless loop
	private LeafNode extractSubtree(TreeNode node, LexEntry lex, boolean first) {
		
		// find the head trunk
		while (true) {
			
			// treat coordination and predicative auxiliary trees
			while (Utils.isCoord(node) || (node.getNodeStatusBottom() == NodeStatus.PREDADJ && first)) {

				TreeNode foot;
				if (node.getNodeStatusBottom() == NodeStatus.PREDADJ) {
					foot = extractPredAdj(node, new LexEntry(node, multiAnchoredTrees));
				}
				else {
					foot = extractCoordination(node);
				}
				if (foot == null) {
					break;
				}
				node = foot;
			}
			
			node.setBottom(lex);
			lex.addNode(node);
			if (!node.isLeaf()) {
				TreeNode head = node.getHeadChild();
				head.setTop(lex);
				lex.addNode(head);
				
				// Recursively also extract the siblings of the head
				TreeNode foot = treatSiblingsOfHead(node, lex);
				if (foot != null) { 
					// there was an adjunction tree in between
					node = foot;
				}
				else {
					node = head;
				}
			}
			else { // anchor found
				return (LeafNode)node;
			}
		}
	}
	

	
	private TreeNode extractCoordination(TreeNode node) {
		LexEntry lex = new LexEntry(node, multiAnchoredTrees);
		lex.setAdjTree();
		node.setBottom(lex);
		lex.addNode(node);		
		
		// this is for ugly coordinated phrases that do not have a "real" head/anchor;
		// we do not finish this current coordination lex entry, but give
		// the node back to the "normal" extraction procedure.
		if (!node.getHeadChild().isLeaf()) {
			return null;
		}
		
		// First run to treat the conjuncts
		TreeNode foot = null;
		int firstCJ = -1;
		int firstCD = -1;
		TreeNode child;
		ArrayList<LexEntry> substs = new ArrayList<LexEntry>();
		for (int i = 0; i < node.getChildren().size(); i++) {
			child = node.getChildren().get(i);
			if (child.getFunctionalCategory().equals("CJ") && !child.isHead()) {
				if (firstCJ == -1) {
					firstCJ = i;
				}
				child.setTop(lex);
				lex.addNode(child);
				if (foot == null && !child.getPhrasalCategory().equals("TRUNC")) { // create foot
					lex.setFoot(child);
					foot = child;
				}
				else { // create substitution site
					child.setNodeStatus(NodeStatus.ARG);
					//extract(child, new LexEntry(child, multiAnchoredTrees));
					LexEntry lexSubst = new LexEntry(child, multiAnchoredTrees);
					LeafNode leaf = extractSubtree(child, lexSubst);
					lexSubst.setLeafLink(leaf);
					substs.add(lexSubst);
				}
			}
			else if (child.getFunctionalCategory().equals("CJ") && child.isHead()) {
				if (firstCJ == -1) {
					firstCJ = i;
				}
				if (foot == null) { // create foot
					//child.setNodeStatus(NodeStatus.FOOT);
					//lex.setFoot(child);
					//foot = child;
				}
			}
			else if (child.getFunctionalCategory().equals("CD")) {
				if (firstCD == -1) {
					firstCD = i;
				}
			}
		}
		
		// if foot==null here, there was no CJ (probably annotation error).
		// we do not extract this as coordination...
		if (foot == null) {
			return null;
		}
		
		// Second run to treat the coordinations
		for (TreeNode child2 : node.getChildren()) {
			if (child2.isHead()) { // the head is not necessarily a CD
				child2.setTop(lex);
				lex.addNode(child2);
				if (child2.isLeaf()) {
					child2.setBottom(lex);
					lex.setLeafLink((LeafNode)child2);
				}
				else {
					// this is for ugly coordinated phrases that do not have a "real" head/anchor;
					// already covered above
				}
			}
			else if (child2.getFunctionalCategory().equals("CD")) {
				if (child2.isHead()) {  // i don't think we need this...
					child2.setTop(lex);
					child2.setBottom(lex);
					lex.addNode(child2);
					lex.setLeafLink((LeafNode)child2);
				}
				else if (firstCD < firstCJ) { // entweder... oder... etc. -> multi-anchored tree
					child2.setNodeStatus(NodeStatus.MAARG);
					child2.setTop(lex);
					lex.addNode(child2);

					LexEntry lexNew = new LexEntry(child2, multiAnchoredTrees);
					//lexNew.addAssociate(lex);
					lex.addAssociate(lexNew);
					//child2.setMATLink(lexNew);
					extract(child2, lexNew);
				}
				else { // substitution site
					child2.setNodeStatus(NodeStatus.ARG);
					child2.setTop(lex);
					lex.addNode(child2);
					extract(child2, new LexEntry(child2, multiAnchoredTrees));
				}
			}
			else if (child2.getFunctionalCategory().equals("CJ") && !child2.isHead()) {
				// we already looked at them before
			}
			else if (Utils.isPunct(child2.getPhrasalCategory())
					&& !Utils.isRightPunctPeriphery(currTigerTree, child2)) { // punctuation as substitution node
				child2.setNodeStatus(NodeStatus.ARG);
				child2.setTop(lex);
				lex.addNode(child2);
				extract(child2, new LexEntry(child2, multiAnchoredTrees));
			}
			else { // modifier
				child2.setNodeStatus(NodeStatus.MOD);
				LexEntry modLex = new LexEntry(child2, multiAnchoredTrees);
				child2.setTop(modLex);
				modLex.addNode(child2);
				extract(child2, modLex);
			}
		}
		currLexicon.addAll(substs);
		currLexicon.add(lex);
		return foot;
	}
	

	private TreeNode treatSiblingsOfHead(TreeNode node, LexEntry lex) {
		
		TreeNode foot = null;
		boolean isAdj = false;
		
		// check whether one of the children is ADJ.
		// if yes, we need a special procedure here.
//		for (TreeNode child : node.getChildren()) {
//			if (!child.isHead() && !child.isFoot()) {
//				NodeStatus stat = child.getNodeStatus();
//				if (stat == NodeStatus.ADJ) {
//					isAdj = true;
//					foot = extractAdj(node, child);
//				}
//			}
//		}
		
		if (!isAdj) {
			for (TreeNode child : node.getChildren()) {
				if (!child.isHead() && !child.isFoot()) {
					NodeStatus stat = child.getNodeStatus();
					if (stat == NodeStatus.ARG) {
						child.setTop(lex);
						lex.addNode(child);
						if (child.isLeaf() && ((LeafNode)child).isTrace()) {
							// bind argument trace into the current lex entry
							addToLexEntry(child, lex);
						}
						else if (currTigerTree.getFoot2Root().containsKey(child.getID())) {
							if (currTigerTree.getFoot2Root().get(child.getID()) == lex.getRoot()) {
								child.isFoot();
								lex.setFoot(child);
							}
							else {
								System.err.println("### Error extracting predicative adj tree " + Utils.subtreeToString(node));
							}
						}
						else {
							extract(child, new LexEntry(child, multiAnchoredTrees));
						}
					}
					else if (stat == NodeStatus.MOD) {
						LexEntry modLex = new LexEntry(child, multiAnchoredTrees);
						modLex.addNode(child);
						child.setTop(modLex);
						extract(child, modLex);
					}
					else if (stat == NodeStatus.MAARG) {
						if (child.isLeaf() && ((LeafNode)child).isTrace()) {
							// bind argument trace into the current lex entry
							addToLexEntry(child, lex);
						}
						else {
							child.setTop(lex);
							lex.addNode(child);
							LexEntry lexNew = new LexEntry(child, multiAnchoredTrees);
							lex.addAssociate(lexNew);
							//child.setMATLink(lexNew);
							extract(child, lexNew);
						}
					}
					else { // should not be needed
						extract(child, new LexEntry(child, multiAnchoredTrees));
					}
				}
			}
		}
		return foot;
	}

//	private TreeNode extractAdj(TreeNode node, TreeNode child) {
//		LexEntry lex = new LexEntry(node, multiAnchoredTrees);
//		lex.setAdjTree();
//		// root
//		node.setBottom(lex);
//		lex.addNode(node);
//		// foot
//		node.getHeadChild().setTop(lex);
//		lex.addNode(node.getHeadChild());
//		//node.getHeadChild().setNodeStatus(NodeStatus.FOOT);
//		node.getHeadChild().setIsFoot();
//		lex.setFoot(node.getHeadChild());
//		// adjunct
//		child.setTop(lex);
//		lex.addNode(child);
//		extract(child, lex);
//		// maybe there are other children as well
//		for (TreeNode child2 : node.getChildren()) {
//			if (!child2.isHead() && !child2.isFoot() && child2 != child) {
//				NodeStatus stat = child2.getNodeStatus();
//				if (stat == NodeStatus.ARG) {
//					child2.setTop(lex);
//					lex.addNode(child2);
//					if (child2.isLeaf() && ((LeafNode)child2).isTrace()) {
//						// bind argument trace into the current lex entry
//						addToLexEntry(child2, lex);
//					}
//					else {
//						extract(child2, new LexEntry(child2, multiAnchoredTrees));
//					}
//				}
//				else if (stat == NodeStatus.MOD) {
//					LexEntry modLex = new LexEntry(child2, multiAnchoredTrees);
//					modLex.addNode(child2);
//					child2.setTop(modLex);
//					extract(child2, modLex);
//				}
//				else if (stat == NodeStatus.MAARG) {
//					child2.setTop(lex);
//					lex.addNode(child2);
//					//if (multiAnchoredTrees) {
//					//	extractMA(child, lex);
//					//}
//					//else { // treat as ARG
//						LexEntry lexNew = new LexEntry(child2, multiAnchoredTrees);
//						child2.setMATLink(lexNew);
//						extract(child2, lexNew);
//					//}
//				}
//				else {
//					extract(child2, new LexEntry(child2, multiAnchoredTrees));
//				}
//			}
//		}
//		
//		return node.getHeadChild();
//	}

}
